# "More than one statement on a single line"
# pylint: disable-msg=C0321

"""
gentoo ebuild atom, should be generalized into an agnostic base
"""

__all__ = ("atom", "transitive_use_atom")

import string

from snakeoil import klass
from snakeoil.compatibility import cmp
from snakeoil.demandload import demand_compile_regexp

from ..restrictions import boolean, packages, restriction, values
from ..restrictions.packages import AndRestriction as PkgAndRestriction
from ..restrictions.packages import Conditional
from ..restrictions.values import ContainmentMatch
from . import cpv, errors, restricts

# namespace compatibility...
MalformedAtom = errors.MalformedAtom

alphanum = set(string.digits)
alphanum.update(string.ascii_letters)

valid_repo_chars = set(alphanum)
valid_repo_chars.update("_-")
valid_slot_chars = set(alphanum)
valid_slot_chars.update(".+_-")
alphanum = frozenset(alphanum)
valid_repo_chars = frozenset(valid_repo_chars)
valid_slot_chars = frozenset(valid_slot_chars)
valid_ops = frozenset(['<', '<=', '=', '~', '>=', '>'])

demand_compile_regexp('valid_use_flag', r'^[A-Za-z0-9][A-Za-z0-9+_@-]*$')


class atom(boolean.AndRestriction, metaclass=klass.generic_equality):
    """Currently implements gentoo ebuild atom parsing.

    Should be converted into an agnostic dependency base.
    """

    # note we don't need _hash
    __slots__ = (
        "blocks", "blocks_strongly", "op", "cpvstr", "negate_vers", "use",
        "slot_operator", "slot", "subslot", "repo_id", "_hash", "_cpv",
        "_restrictions",
    )

    type = restriction.package_type

    negate = False

    _evaluate_collapse = True

    __attr_comparison__ = (
        "cpvstr", "op", "blocks", "negate_vers",
        "use", "slot", "subslot", "slot_operator", "repo_id")

    klass.inject_richcmp_methods_from_cmp(locals())
    # hack; combine these 2 metaclasses at some point...
    locals().pop("__eq__", None)
    locals().pop("__ne__", None)
    __inst_caching__ = True

    # overrided in child class if it's supported
    evaluate_depset = None

    def __init__(self, atom, negate_vers=False, eapi='-1'):
        """
        :param atom: string, see gentoo ebuild atom syntax
        :keyword negate_vers: boolean controlling whether the version should be
            inverted for restriction matching
        :keyword eapi: string/int controlling what eapi to enforce for this atom
        """
        if not atom:
            raise errors.MalformedAtom(atom)

        sf = object.__setattr__
        orig_atom = atom
        override_kls = False
        use_start = atom.find("[")
        slot_start = atom.find(":")

        if use_start != -1:
            # use dep
            use_end = atom.find("]", use_start)
            if use_end == -1:
                raise errors.MalformedAtom(
                    orig_atom, "use restriction isn't completed")
            elif use_end != len(atom) - 1:
                raise errors.MalformedAtom(
                    orig_atom, "trailing garbage after use dep")
            sf(self, "use", tuple(sorted(atom[use_start + 1:use_end].split(','))))
            for x in self.use:
                # stripped purely for validation reasons
                try:
                    if x[-1] in "=?":
                        override_kls = True
                        x = x[:-1]
                        if x[0] == '!':
                            x = x[1:]
                        if x[0] == '-':
                            raise errors.MalformedAtom(
                                orig_atom, f"malformed use flag: {x}")
                    elif x[0] == '-':
                        x = x[1:]

                    if x[-1] == ')' and eapi not in ('0', '1', '2', '3'):
                        # use defaults.
                        if x[-3:] in ("(+)", "(-)"):
                            x = x[:-3]

                    if not x:
                        raise errors.MalformedAtom(orig_atom, 'empty use dep detected')
                    if not valid_use_flag.match(x):
                        raise errors.MalformedAtom(orig_atom, f'invalid USE flag: {x!r}')
                except IndexError:
                    raise errors.MalformedAtom(
                        orig_atom, 'empty use dep detected')
            if override_kls:
                sf(self, '__class__', transitive_use_atom)
            atom = atom[0:use_start]+atom[use_end + 1:]
        else:
            sf(self, "use", None)
        if slot_start != -1:
            i2 = atom.find("::", slot_start)
            if i2 != -1:
                repo_id = atom[i2 + 2:]
                if not repo_id:
                    raise errors.MalformedAtom(
                        orig_atom, "repo_id must not be empty")
                elif repo_id[0] in '-':
                    raise errors.MalformedAtom(
                        orig_atom,
                        f"invalid first char of repo_id '{repo_id}' (must not begin with a hyphen)")
                elif not valid_repo_chars.issuperset(repo_id):
                    raise errors.MalformedAtom(
                        orig_atom,
                        f"repo_id may contain only [a-Z0-9_-/], found {repo_id!r}")
                atom = atom[:i2]
                sf(self, "repo_id", repo_id)
            else:
                sf(self, "repo_id", None)
            # slot dep.
            slot = atom[slot_start+1:]
            slot_operator = subslot = None
            if not slot:
                # if the slot char came in only due to repo_id, force slots to None
                if i2 == -1:
                    raise errors.MalformedAtom(
                        orig_atom, "Empty slot targets aren't allowed")
                slot = None
            else:
                slots = (slot,)
                if eapi not in ('0', '1', '2', '3', '4'):
                    if slot[0:1] in ("*", "="):
                        if len(slot) > 1:
                            raise errors.MalformedAtom(
                                orig_atom,
                                "Slot operators '*' and '=' do not take slot targets")
                        slot_operator = slot
                        slot, slots = None, ()
                    else:
                        if slot.endswith('='):
                            slot_operator = '='
                            slot = slot[:-1]
                        slots = slot.split('/', 1)
                elif eapi == '0':
                    raise errors.MalformedAtom(
                        orig_atom,
                        "slot dependencies aren't allowed in EAPI 0")

                for chunk in slots:
                    if not chunk:
                        raise errors.MalformedAtom(
                            orig_atom,
                            "Empty slot targets aren't allowed")

                    if chunk[0] in '-.':
                        raise errors.MalformedAtom(
                            orig_atom,
                            "Slot targets must not start with a hypen or dot: {chunk!r}")
                    elif not valid_slot_chars.issuperset(chunk):
                        invalid_chars = ', '.join(map(repr, sorted(set(chunk).difference(valid_slot_chars))))
                        raise errors.MalformedAtom(
                            orig_atom,
                            f"Invalid character(s) in slot target: {invalid_chars}")

                if len(slots) == 2:
                    slot, subslot = slots

            sf(self, "slot_operator", slot_operator)
            sf(self, "slot", slot)
            sf(self, "subslot", subslot)
            atom = atom[:slot_start]
        else:
            sf(self, "slot_operator", None)
            sf(self, "slot", None)
            sf(self, "subslot", None)
            sf(self, "repo_id", None)

        sf(self, "blocks", atom[0] == "!")
        if self.blocks:
            atom = atom[1:]
            # hackish/slow, but lstrip doesn't take a 'prune this many' arg
            # open to alternatives
            if eapi not in ('0', '1') and atom.startswith("!"):
                atom = atom[1:]
                sf(self, "blocks_strongly", True)
            else:
                sf(self, "blocks_strongly", False)
        else:
            sf(self, "blocks_strongly", False)

        if atom[0] in ('<', '>'):
            if atom[1] == '=':
                sf(self, 'op', atom[:2])
                atom = atom[2:]
            else:
                sf(self, 'op', atom[0])
                atom = atom[1:]
        elif atom[0] == '=':
            if atom[-1] == '*':
                sf(self, 'op', '=*')
                atom = atom[1:-1]
            else:
                atom = atom[1:]
                sf(self, 'op', '=')
        elif atom[0] == '~':
            sf(self, 'op', '~')
            atom = atom[1:]
        else:
            sf(self, 'op', '')
        sf(self, 'cpvstr', atom)

        if eapi == '0':
            for x in ('use', 'slot'):
                if getattr(self, x) is not None:
                    raise errors.MalformedAtom(
                        orig_atom,
                        f"{x} atoms aren't supported for EAPI 0")
        elif eapi == '1':
            if self.use is not None:
                raise errors.MalformedAtom(
                    orig_atom,
                    "use atoms aren't supported for EAPI < 2")
        if eapi != '-1':
            if self.repo_id is not None:
                raise errors.MalformedAtom(
                    orig_atom,
                    f"repo_id atoms aren't supported for EAPI {eapi}")
        if use_start != -1 and slot_start != -1 and use_start < slot_start:
            raise errors.MalformedAtom(
                orig_atom,
                "slot restriction must proceed use")
        try:
            sf(self, "_cpv", cpv.CPV(self.cpvstr, versioned=bool(self.op)))
        except errors.InvalidCPV as e:
            raise errors.MalformedAtom(orig_atom) from e

        if self.op:
            if self.version is None:
                raise errors.MalformedAtom(
                    orig_atom, "operator requires a version")
            elif self.op == '~' and self.revision:
                raise errors.MalformedAtom(
                    orig_atom,
                    "~ revision operater cannot be combined with a revision")
        elif self.version is not None:
            raise errors.MalformedAtom(
                orig_atom, 'versioned atom requires an operator')
        sf(self, "_hash", hash(orig_atom))
        sf(self, "negate_vers", negate_vers)

    __getattr__ = klass.GetAttrProxy("_cpv")
    __dir__ = klass.DirProxy("_cpv")

    @property
    def blocks_temp_ignorable(self):
        return not self.blocks_strongly

    weak_blocker = klass.alias_attr("blocks_temp_ignorable")

    def __repr__(self):
        if self.op == '=*':
            atom = f"={self.cpvstr}*"
        else:
            atom = self.op + self.cpvstr
        if self.blocks:
            atom = '!' + atom
        if self.blocks:
            if self.blocks_strongly:
                atom = '!!' + atom
            else:
                atom = '!' + atom
        attrs = [atom]
        if self.use:
            attrs.append(f'use={self.use!r}')
        if self.slot is not None:
            attrs.append(f'slot={self.slot!r}')
        if self.subslot is not None:
            attrs.append(f'subslot={self.subslot!r}')
        if self.repo_id is not None:
            attrs.append(f'repo_id={self.repo_id!r}')
        return '<%s %s @#%x>' % (
            self.__class__.__name__, ' '.join(attrs), id(self))

    def __reduce__(self):
        return (atom, (str(self), self.negate_vers))

    def iter_dnf_solutions(self, full_solution_expansion=False):
        if full_solution_expansion:
            return boolean.AndRestriction.iter_dnf_solutions(self, True)
        return iter([[self]])

    def iter_cnf_solutions(self, full_solution_expansion=False):
        if full_solution_expansion:
            return boolean.AndRestriction.iter_cnf_solutions(self, True)
        return iter([[self]])

    def cnf_solutions(self, full_solution_expansion=False):
        if full_solution_expansion:
            return boolean.AndRestriction.cnf_solutions(self, True)
        return [[self]]

    @property
    def is_simple(self):
        return len(self.restrictions) == 2

    @klass.jit_attr
    def restrictions(self):
        # ordering here matters; against 24702 ebuilds for
        # a non matchable atom with package as the first restriction
        # 10 loops, best of 3: 206 msec per loop
        # with category as the first(the obvious ordering)
        # 10 loops, best of 3: 209 msec per loop
        # why?  because category is more likely to collide;
        # at the time of this profiling, there were 151 categories.
        # over 11k packages however.
        r = [restricts.PackageDep(self.package), restricts.CategoryDep(self.category)]

        if self.repo_id is not None:
            r.insert(0, restricts.RepositoryDep(self.repo_id))

        if self.fullver is not None:
            if self.op == '=*':
                r.append(packages.PackageRestriction(
                    "fullver", values.StrGlobMatch(self.fullver)))
            else:
                r.append(restricts.VersionMatch(
                    self.op, self.version, self.revision, negate=self.negate_vers))

        if self.slot is not None:
            r.append(restricts.SlotDep(self.slot))
            if self.subslot is not None:
                r.append(restricts.SubSlotDep(self.subslot))

        if self.use is not None:
            r.extend(restricts._parse_nontransitive_use(self.use))

        return tuple(r)

    def __str__(self):
        if self.op == '=*':
            s = f"={self.cpvstr}*"
        else:
            s = self.op + self.cpvstr
        if self.blocks:
            if self.blocks_strongly:
                s = '!!' + s
            else:
                s = '!' + s
        if self.slot:
            s += f":{self.slot}"
            if self.subslot:
                s += f"/{self.subslot}"
            if self.slot_operator == "=":
                s += self.slot_operator
        elif self.slot_operator:
            s += f":{self.slot_operator}"
        if self.repo_id:
            s += f"::{self.repo_id}"
        if self.use:
            use = ','.join(self.use)
            s += f"[{use}]"
        return s

    __hash__ = klass.reflective_hash('_hash')

    def __iter__(self):
        return iter(self.restrictions)

    def __getitem__(self, index):
        return self.restrictions[index]

    def __cmp__(self, other):
        if not isinstance(other, atom):
            raise TypeError(
                f"other isn't of {atom!r} type, is {other.__class__}")

        c = cmp(self.category, other.category)
        if c:
            return c

        c = cmp(self.package, other.package)
        if c:
            return c

        c = cmp(self.op, other.op)
        if c:
            return c

        c = cpv.ver_cmp(self.version, self.revision,
                        other.version, other.revision)
        if c:
            return c

        c = cmp(self.blocks, other.blocks)
        if c:
            # invert it; cmp(True, False) == 1
            # want non blockers then blockers.
            return -c

        c = cmp(self.blocks_strongly, other.blocks_strongly)
        if c:
            # want !! prior to !
            return c

        c = cmp(self.negate_vers, other.negate_vers)
        if c:
            return c

        def f(v):
            return '' if v is None else v
        c = cmp(f(self.slot), f(other.slot))
        if c:
            return c

        c = cmp(self.use, other.use)
        if c:
            return c

        return cmp(self.repo_id, other.repo_id)

    no_usedeps = klass.alias_attr("get_atom_without_use_deps")

    @property
    def get_atom_without_use_deps(self):
        """Return atom object stripped of USE dependencies."""
        if not self.use:
            return self
        if self.op == '=*':
            s = f'={self.cpvstr}*'
        else:
            s = self.op + self.cpvstr
        if self.blocks:
            s = '!' + s
            if not self.blocks_temp_ignorable:
                s = '!' + s
        if self.slot:
            s += f':{self.slot}'
        return atom(s)

    def intersects(self, other):
        """Check if a passed in atom "intersects" this restriction's atom.

        Two atoms "intersect" if a package can be constructed that
        matches both:

        - if you query for just "dev-lang/python" it "intersects" both
          "dev-lang/python" and ">=dev-lang/python-2.4"
        - if you query for "=dev-lang/python-2.4" it "intersects"
          ">=dev-lang/python-2.4" and "dev-lang/python" but not
          "<dev-lang/python-2.3"

        USE and slot deps are also taken into account.

        The block/nonblock state of the atom is ignored.
        """
        # Our "key" (cat/pkg) must match exactly:
        if self.key != other.key:
            return False

        # Slot dep only matters if we both have one. If we do they
        # must be identical:
        if (self.slot is not None and other.slot is not None and
                self.slot != other.slot):
            return False

        # Subslot dep only matters if we both have one. If we do they
        # must be identical:
        if (self.subslot is not None and other.subslot is not None and
                self.subslot != other.subslot):
            return False

        if (self.repo_id is not None and other.repo_id is not None and
                self.repo_id != other.repo_id):
            return False

        # Use deps are similar: if one of us forces a flag on and the
        # other forces it off we do not intersect. If only one of us
        # cares about a flag it is irrelevant.

        # Skip the (very common) case of one of us not having use deps:
        if self.use and other.use:
            # Set of flags we do not have in common:
            flags = set(self.use) ^ set(other.use)
            for flag in flags:
                # If this is unset and we also have the set version we fail:
                if flag[0] == '-' and flag[1:] in flags:
                    return False

        # Remaining thing to check is version restrictions. Get the
        # ones we can check without actual version comparisons out of
        # the way first.

        # If one of us is unversioned we intersect:
        if not self.op or not other.op:
            return True

        # If we are both "unbounded" in the same direction we intersect:
        if (('<' in self.op and '<' in other.op) or
                ('>' in self.op and '>' in other.op)):
            return True

        # Trick used here: just use the atoms as sufficiently
        # package-like object to pass to these functions (all that is
        # needed is a version and revision attr).

        # If one of us is an exact match we intersect if the other matches it:
        if self.op == '=':
            if other.op == '=*':
                return self.fullver.startswith(other.fullver)
            return restricts.VersionMatch(
                other.op, other.version, other.revision).match(self)
        if other.op == '=':
            if self.op == '=*':
                return other.fullver.startswith(self.fullver)
            return restricts.VersionMatch(
                self.op, self.version, self.revision).match(other)

        # If we are both ~ matches we match if we are identical:
        if self.op == other.op == '~':
            return (self.version == other.version and
                    self.revision == other.revision)

        # If we are both glob matches we match if one of us matches the other.
        if self.op == other.op == '=*':
            return (self.fullver.startswith(other.fullver) or
                    other.fullver.startswith(self.fullver))

        # If one of us is a glob match and the other a ~ we match if the glob
        # matches the ~ (ignoring a revision on the glob):
        if self.op == '=*' and other.op == '~':
            return other.fullver.startswith(self.version)
        if other.op == '=*' and self.op == '~':
            return self.fullver.startswith(other.version)

        # If we get here at least one of us is a <, <=, > or >=:
        if self.op in ('<', '<=', '>', '>='):
            ranged, other = self, other
        else:
            ranged, other = other, self

        if '<' in other.op or '>' in other.op:
            # We are both ranged, and in the opposite "direction" (or
            # we would have matched above). We intersect if we both
            # match the other's endpoint (just checking one endpoint
            # is not enough, it would give a false positive on <=2 vs >2)
            return (
                restricts.VersionMatch(
                    other.op, other.version, other.revision).match(ranged) and
                restricts.VersionMatch(
                    ranged.op, ranged.version, ranged.revision).match(other))

        if other.op == '~':
            # Other definitely matches its own version. If ranged also
            # does we're done:
            if restricts.VersionMatch(
                    ranged.op, ranged.version, ranged.revision).match(other):
                return True
            # The only other case where we intersect is if ranged is a
            # > or >= on other's version and a nonzero revision. In
            # that case other will match ranged. Be careful not to
            # give a false positive for ~2 vs <2 here:
            return ranged.op in ('>', '>=') and restricts.VersionMatch(
                other.op, other.version, other.revision).match(ranged)

        if other.op == '=*':
            # The fun one, since glob matches do not correspond to a
            # single contiguous region of versions.

            # a glob match definitely matches its own version, so if
            # ranged does too we're done:
            if restricts.VersionMatch(
                    ranged.op, ranged.version, ranged.revision).match(other):
                return True
            if '<' in ranged.op:
                # Remaining cases where this intersects: there is a
                # package smaller than ranged.fullver and
                # other.fullver that they both match.

                # If other.revision is not None or 0 then other does not match
                # anything smaller than its own fullver:
                if other.revision:
                    return False

                # If other.revision is None or 0 then we can always
                # construct a package smaller than other.fullver by
                # tagging e.g. an _alpha1 on, since
                # cat/pkg_beta2_alpha1_alpha1 is a valid version.
                # (Yes, really. Try it if you don't believe me.)
                # If and only if other also matches ranged then
                # ranged will also match one of those smaller packages.
                # XXX (I think, need to try harder to verify this.)
                return ranged.fullver.startswith(other.version)
            else:
                # Remaining cases where this intersects: there is a
                # package greater than ranged.fullver and
                # other.fullver that they both match.

                # We can always construct a package greater than
                # other.fullver by adding a digit to it.
                # If and only if other also matches ranged then
                # ranged will match such a larger package
                # XXX (I think, need to try harder to verify this.)
                return ranged.fullver.startswith(other.version)

        # Handled all possible ops.
        raise NotImplementedError(
            'Someone added an op to atom without adding it to intersects')

    def evaluate_conditionals(self, parent_cls, parent_seq, enabled, tristate=None):
        parent_seq.append(self)


class transitive_use_atom(atom):

    __slots__ = ()
    __inst_caching__ = True
    _nontransitive_use_atom = atom

    is_simple = False

    def _stripped_use(self):
        return str(self).split("[", 1)[0]

    @staticmethod
    def _mk_conditional(flag, payload, negate=False):
        return Conditional('use', ContainmentMatch(flag, negate=negate), payload)

    def _recurse_transitive_use_conds(self, atom_str, forced_use, varied):
        if not varied:
            s = ','.join(forced_use)
            if s:
                s = f'[{s}]'
            return (self._nontransitive_use_atom(atom_str + s), )

        flag = varied[0]
        use = flag.lstrip('!').rstrip('?=')
        varied = varied[1:]
        if flag[-1] == '?':
            # a[x?] == x? ( a[x] ) !x? ( a )
            # a[!x?] == x? ( a ) !x? ( a[-x] )
            if flag[0] != '!':
                return (self._mk_conditional(use,
                        self._recurse_transitive_use_conds(
                            atom_str, forced_use + [use], varied)),
                        self._mk_conditional(
                            use, self._recurse_transitive_use_conds(
                                atom_str, forced_use, varied), negate=True))
            return (self._mk_conditional(
                    use, self._recurse_transitive_use_conds(
                        atom_str, forced_use, varied)),
                    self._mk_conditional(
                        use, self._recurse_transitive_use_conds(
                            atom_str, forced_use + ['-' + use], varied), negate=True))

        # a[x=] == x? ( a[x] ) !x? ( a[-x] )
        # a[!x=] == x? ( a[-x] ) !x? ( a[x] )
        if flag[0] != '!':
            use_states = [[use], ['-' + use]]
        else:
            use_states = [['-' + use], [use]]

        return (self._mk_conditional(
                use, self._recurse_transitive_use_conds(
                    atom_str, forced_use + use_states[0], varied)),
                self._mk_conditional(
                    use, self._recurse_transitive_use_conds(
                        atom_str, forced_use + use_states[1], varied), negate=True))

    @klass.jit_attr
    def restrictions(self):
        return self.convert_to_conditionals()

    def convert_to_conditionals(self):
        static_use = [use for use in self.use if use[-1] not in '?=']
        variable = [use for use in self.use if use[-1] in '?=']
        return PkgAndRestriction(
            *self._recurse_transitive_use_conds(
                self._stripped_use(), static_use, variable))

    def _evaluate_depset_qa_in_place(self, flags, variable_flags, enabled, tristate):
        # note this mutates flags
        for flag in variable_flags:
            conditional = flag[-1]
            negated = flag[0] == '!'
            if negated:
                flag = flag[1:-1]
            else:
                flag = flag[:-1]

            real_flag = flag
            if flag[-1] == ')':
                flag = flag[:-3]

            if conditional == '=':
                # if it's locked to a state, take that state; else use whatever
                # the default state was.
                if flag in tristate:
                    # it's locked; take the allowed state, and force it, take that bool
                    # and use it to decide if the target dep gets a disabled assertion.
                    # roughly if locked to enabled: x= == x, !x= == -x
                    # if locked to disabled: x= == -x , !x= == x
                    if (flag in enabled) == negated:
                        real_flag = '-' + real_flag
            else:
                if flag in tristate:
                    # if the flag was on, but it was !x?, then skip it.
                    # if the flag was off, but it was x?, then skip it.
                    if (flag in enabled) == negated:
                        continue
                    # enforce the allowed state.
                    if flag not in enabled:
                        real_flag = '-' + real_flag
                else:
                    # enforce the state that gets us a flag to test on the target.
                    # thus if !x?, we want -x, or +x; take the negation basically.
                    if negated:
                        real_flag = '-' + real_flag
            flags.append(real_flag)

    def evaluate_conditionals(self, parent_cls, parent_seq, enabled, tristate_filter=None):
        new_flags = [use for use in self.use if use[-1] not in '?=']
        variable_flags = [use for use in self.use if use[-1] in '?=']

        if tristate_filter is not None:
            # note this updates the flags in place.
            self._evaluate_depset_qa_in_place(
                new_flags, variable_flags, enabled, tristate_filter)
        else:
            for flag in variable_flags:
                conditional = flag[-1]
                negated = flag[0] == '!'
                if negated:
                    flag = raw_flag = flag[1:-1]
                else:
                    flag = raw_flag = flag[:-1]

                if raw_flag[-1] == ')':
                    # use default... strip "(+)"
                    raw_flag = raw_flag[:-3]

                if conditional == '=':
                    # given '!x=', if x is off, force x on for the target,
                    # and vice versa.  render out a non relative - or ''.
                    negated = ((raw_flag in enabled) == negated)
                    if negated:
                        flag = '-' + flag
                else:
                    # enforce the flag only if our state matches.  !x? and x is on, means no dep.
                    # for !x? with -x, the assertion becomes !x; conditionally transitive basically.
                    if (raw_flag in enabled) == negated:
                        continue
                    flag = "%s%s" % (negated and "-" or "", flag)
                new_flags.append(flag)

        if not new_flags:
            a = self._nontransitive_use_atom(self._stripped_use())
        else:
            a = self._nontransitive_use_atom(
                "%s[%s]" % (self._stripped_use(), ','.join(new_flags)))
        parent_seq.append(a)

    iter_dnf_solutions = boolean.AndRestriction.iter_dnf_solutions
    cnf_solutions = boolean.AndRestriction.cnf_solutions



"""
gentoo configuration domain
"""

__all__ = ("domain",)

# XXX doc this up better...

import copy
import os
import re
import tempfile
from collections import defaultdict
from functools import partial
from itertools import chain
from multiprocessing import cpu_count
from operator import itemgetter

from snakeoil import klass
from snakeoil.bash import iter_read_bash, read_bash_dict
from snakeoil.cli.exceptions import find_user_exception
from snakeoil.data_source import local_source
from snakeoil.log import suppress_logging
from snakeoil.mappings import ImmutableDict, ProtectedDict
from snakeoil.osutils import pjoin
from snakeoil.process.spawn import spawn_get_output
from snakeoil.sequences import predicate_split, split_negations, stable_unique

from ..binpkg import repository as binary_repo
from ..cache.flat_hash import md5_cache
from ..config import basics
from ..config import errors as config_errors
from ..config.domain import Failure
from ..config.domain import domain as config_domain
from ..config.hint import ConfigHint
from ..fs.livefs import iter_scan, sorted_scan
from ..log import logger
from ..repository import errors as repo_errors
from ..repository import filtered
from ..repository.util import RepositoryGroup
from ..restrictions import packages, values
from ..restrictions.delegated import delegate
from ..util.parserestrict import ParseError, parse_match
from . import const
from . import repository as ebuild_repo
from .atom import atom as _atom
from .misc import (ChunkedDataDict, chunked_data, collapsed_restrict_to_data, incremental_expansion,
                   incremental_expansion_license, non_incremental_collapsed_restrict_to_data,
                   optimize_incrementals)
from .portage_conf import PortageConfig
from .repo_objs import Licenses, RepoConfig
from .triggers import GenerateTriggers


def package_masks(iterable):
    for line, lineno, path in iterable:
        try:
            yield parse_match(line), line, lineno, path
        except ParseError as e:
            logger.warning(f'{path!r}, line {lineno}: parsing error: {e}')


def package_keywords_splitter(iterable):
    for line, lineno, path in iterable:
        v = line.split()
        try:
            yield parse_match(v[0]), tuple(v[1:]), line, lineno, path
        except ParseError as e:
            logger.warning(f'{path!r}, line {lineno}: parsing error: {e}')


def package_env_splitter(basedir, iterable):
    for line, lineno, path in iterable:
        val = line.split()
        if len(val) == 1:
            logger.warning(f"{path!r}, line {lineno}: missing file reference: {line!r}")
            continue
        paths = []
        for env_file in val[1:]:
            fp = pjoin(basedir, env_file)
            if os.path.exists(fp):
                paths.append(fp)
            else:
                logger.warning(f"{path!r}, line {lineno}: nonexistent file: {fp!r}")
        try:
            yield parse_match(val[0]), tuple(paths), line, lineno, path
        except ParseError as e:
            logger.warning(f'{path!r}, line {lineno}: parsing error: {e}')


def apply_mask_filter(globs, atoms, pkg, mode):
    # mode is ignored; non applicable.
    for r in chain(globs, atoms.get(pkg.key, ())):
        if r.match(pkg):
            return True
    return False


def make_mask_filter(masks, negate=False):
    atoms = defaultdict(list)
    globs = []
    for m in masks:
        if isinstance(m, _atom):
            atoms[m.key].append(m)
        else:
            globs.append(m)
    return delegate(partial(apply_mask_filter, globs, atoms), negate=negate)


def generate_filter(masks, unmasks, *extra):
    # note that we ignore unmasking if masking isn't specified.
    # no point, mainly
    masking = make_mask_filter(masks, negate=True)
    unmasking = make_mask_filter(unmasks, negate=False)
    r = ()
    if masking:
        if unmasking:
            r = (packages.OrRestriction(masking, unmasking, disable_inst_caching=True),)
        else:
            r = (masking,)
    return packages.AndRestriction(disable_inst_caching=True, finalize=True, *(r + extra))


def _read_config_file(path):
    """Read all the data files under a given path."""
    try:
        for fs_obj in iter_scan(path, follow_symlinks=True):
            if not fs_obj.is_reg or '/.' in fs_obj.location:
                continue
            for lineno, line in iter_read_bash(
                    fs_obj.location, allow_line_cont=True, enum_line=True):
                yield line, lineno, fs_obj.location
    except FileNotFoundError:
        pass
    except EnvironmentError as e:
        raise Failure(f"failed reading {path!r}: {e}") from e


def load_property(filename, *, read_func=_read_config_file,
                  parse_func=lambda x: x, fallback=()):
    """Decorator for parsing files using specified read/parse methods.

    :param filename: The filename to parse within the config directory.
    :keyword read_func: An invokable used to read the specified file.
    :keyword parse_func: An invokable used to parse the data.
    :keyword fallback: What to return if the file does not exist.
    :return: A :py:`klass.jit.attr_named` property instance.
    """
    def f(func):
        def _load_and_invoke(func, fallback, self, *args, **kwargs):
            if filename.startswith(os.path.sep):
                # translate root fs calls to prefixed root fs
                path = pjoin(self.root, filename.lstrip(os.path.sep))
            else:
                # assume relative files are inside the config dir
                path = pjoin(self.config_dir, filename)
            if os.path.exists(path):
                data = parse_func(read_func(path))
            else:
                data = fallback
            return func(self, data, *args, **kwargs)
        doc = getattr(func, '__doc__', None)
        jit_attr_named = klass.jit_attr_named(f'_jit_{func.__name__}', doc=doc)
        return jit_attr_named(partial(_load_and_invoke, func, fallback))
    return f


# ow ow ow ow ow ow....
# this manages a *lot* of crap.  so... this is fun.
#
# note also, that this is rather ebuild centric. it shouldn't be, and
# should be redesigned to be a seperation of configuration
# instantiation manglers, and then the ebuild specific chunk (which is
# selected by config)
class domain(config_domain):

    # XXX ouch, verify this crap and add defaults and stuff
    _types = {'profile': 'ref:profile', 'repos': 'lazy_refs:repo', 'vdb': 'lazy_refs:repo'}
    for _thing in (
            'root', 'config_dir', 'CHOST', 'CBUILD', 'CTARGET', 'CFLAGS',
            'PATH', 'PORTAGE_TMPDIR', 'DISTCC_PATH', 'DISTCC_DIR', 'CCACHE_DIR'):
        _types[_thing] = 'str'

    # TODO this is missing defaults
    pkgcore_config_type = ConfigHint(
        _types, typename='domain',
        required=['repos', 'profile', 'vdb'],
        allow_unknowns=True)

    del _types, _thing

    def __init__(self, profile, repos, vdb, root='/', prefix='/',
                 config_dir='/etc/portage', **settings):
        self.root = settings["ROOT"] = root
        self.config_dir = config_dir
        self.prefix = prefix
        self.ebuild_hook_dir = pjoin(self.config_dir, 'env')
        self.profile = profile
        self.__repos = repos
        self.__vdb = vdb

        # prevent critical variables from being changed in make.conf
        for k in self.profile.profile_only_variables.intersection(settings.keys()):
            del settings[k]

        # Protect original settings from being overridden so matching
        # package.env settings can be overlaid properly.
        self._settings = ProtectedDict(settings)

    @load_property("/etc/profile.env", read_func=read_bash_dict)
    def system_profile(self, data):
        # prepend system profile $PATH if it exists
        if 'PATH' in data:
            path = stable_unique(
                data['PATH'].split(os.pathsep) + os.environ['PATH'].split(os.pathsep))
            os.environ['PATH'] = os.pathsep.join(path)
        return ImmutableDict(data)

    @klass.jit_attr_named('_jit_reset_settings', uncached_val=None)
    def settings(self):
        settings = self._settings
        if 'CHOST' in settings and 'CBUILD' not in settings:
            settings['CBUILD'] = settings['CHOST']

        # if unset, MAKEOPTS defaults to CPU thread count
        if 'MAKEOPTS' not in settings:
            settings['MAKEOPTS'] = '-j%i' % cpu_count()

        # reformat env.d and make.conf incrementals
        system_profile_settings = {}
        for x in const.incrementals:
            system_profile_val = self.system_profile.get(x, ())
            make_conf_val = settings.get(x, ())
            if isinstance(system_profile_val, str):
                system_profile_val = tuple(system_profile_val.split())
            if isinstance(make_conf_val, str):
                make_conf_val = tuple(make_conf_val.split())
            system_profile_settings[x] = system_profile_val
            settings[x] = make_conf_val

        # roughly... all incremental stacks should be interpreted left -> right
        # as such we start with the env.d settings, append profile settings,
        # and finally append make.conf settings onto that.
        for k, v in self.profile.default_env.items():
            if k not in settings:
                settings[k] = v
                continue
            if k in const.incrementals:
                settings[k] = system_profile_settings[k] + v + settings[k]

        # next we finalize incrementals.
        for incremental in const.incrementals:
            # Skip USE/ACCEPT_LICENSE for the time being; hack; we need the
            # negations currently so that pkg iuse induced enablings can be
            # disabled by negations. For example, think of the profile doing
            # USE=-cdr for brasero w/ IUSE=+cdr. Similarly, ACCEPT_LICENSE is
            # skipped because negations are required for license filtering.
            if incremental not in settings or incremental in ("USE", "ACCEPT_LICENSE"):
                continue
            settings[incremental] = tuple(incremental_expansion(
                settings[incremental],
                msg_prefix=f'while expanding {incremental}'))

        if 'ACCEPT_KEYWORDS' not in settings:
            raise Failure("No ACCEPT_KEYWORDS setting detected from profile, "
                          "or user config")
        settings['ACCEPT_KEYWORDS'] = incremental_expansion(
            settings['ACCEPT_KEYWORDS'],
            msg_prefix='while expanding ACCEPT_KEYWORDS')

        # pull trigger options from the env
        self._triggers = GenerateTriggers(self, settings)

        return ImmutableDict(settings)

    @property
    def arch(self):
        try:
            return self.settings['ARCH']
        except KeyError:
            raise Failure("No ARCH setting detected from profile, or user config")

    @property
    def distdir(self):
        try:
            return self.settings['DISTDIR']
        except KeyError:
            raise Failure("No DISTDIR setting detected from config")

    @property
    def stable_arch(self):
        return self.arch

    @property
    def unstable_arch(self):
        return f"~{self.arch}"

    @klass.jit_attr_named('_jit_reset_features', uncached_val=None)
    def features(self):
        conf_features = list(self.settings.get('FEATURES', ()))
        env_features = os.environ.get('FEATURES', '').split()
        return frozenset(optimize_incrementals(conf_features + env_features))

    @klass.jit_attr_named('_jit_reset_use', uncached_val=None)
    def use(self):
        # append expanded use, FEATURES, and environment defined USE flags
        use = list(self.settings.get('USE', ())) + list(self.profile.expand_use(self.settings))

        # hackish implementation; if test is on, flip on the flag
        if "test" in self.features:
            use.append("test")
        if "prefix" in self.features:
            use.append("prefix")

        return frozenset(optimize_incrementals(use + os.environ.get('USE', '').split()))

    @klass.jit_attr_named('_jit_reset_enabled_use', uncached_val=None)
    def enabled_use(self):
        use = ChunkedDataDict()
        use.add_bare_global(*split_negations(self.use))
        use.merge(self.profile.pkg_use)
        use.update_from_stream(chunked_data(k, *v) for k, v in self.pkg_use)
        use.freeze()
        return use

    @klass.jit_attr_none
    def forced_use(self):
        use = ChunkedDataDict()
        use.merge(getattr(self.profile, 'forced_use'))
        use.add_bare_global((), (self.arch,))
        use.freeze()
        return use

    @klass.jit_attr_none
    def stable_forced_use(self):
        use = ChunkedDataDict()
        use.merge(getattr(self.profile, 'stable_forced_use'))
        use.add_bare_global((), (self.arch,))
        use.freeze()
        return use

    @load_property("package.mask", parse_func=package_masks)
    def pkg_masks(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple(x[0] for x in data)

    @load_property("package.unmask", parse_func=package_masks)
    def pkg_unmasks(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple(x[0] for x in data)

    # TODO: deprecated, remove in 0.11
    @load_property("package.keywords", parse_func=package_keywords_splitter)
    def pkg_keywords(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple((x[0], stable_unique(x[1])) for x in data)

    @load_property("package.accept_keywords", parse_func=package_keywords_splitter)
    def pkg_accept_keywords(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple((x[0], stable_unique(x[1])) for x in data)

    @load_property("package.license", parse_func=package_keywords_splitter)
    def pkg_licenses(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple((x[0], stable_unique(x[1])) for x in data)

    @load_property("package.use", parse_func=package_keywords_splitter)
    def pkg_use(self, data, debug=False):
        if debug:
            return tuple(data)
        return tuple((x[0], split_negations(stable_unique(x[1]))) for x in data)

    @load_property("package.env")
    def pkg_env(self, data, debug=False):
        func = partial(package_env_splitter, self.ebuild_hook_dir)
        data = func(data)
        if debug:
            return tuple(data)
        return tuple((x[0], x[1]) for x in data)

    @klass.jit_attr
    def bashrcs(self):
        files = sorted_scan(pjoin(self.config_dir, 'bashrc'), follow_symlinks=True)
        return tuple(local_source(x) for x in files)

    def _pkg_filters(self, pkg_accept_keywords=None, pkg_keywords=None):
        if pkg_accept_keywords is None:
            pkg_accept_keywords = self.pkg_accept_keywords
        if pkg_keywords is None:
            pkg_keywords = self.pkg_keywords

        # ~amd64 -> [amd64, ~amd64]
        default_keywords = set([self.arch])
        default_keywords.update(self.settings['ACCEPT_KEYWORDS'])
        for x in self.settings['ACCEPT_KEYWORDS']:
            if x.startswith("~"):
                default_keywords.add(x.lstrip("~"))

        # create keyword filters
        accept_keywords = (
            pkg_keywords + pkg_accept_keywords + self.profile.accept_keywords)
        filters = [self._make_keywords_filter(
            default_keywords, accept_keywords,
            incremental="package.keywords" in const.incrementals)]

        # add license filters
        master_license = []
        master_license.extend(self.settings.get('ACCEPT_LICENSE', ()))
        if master_license or self.pkg_licenses:
            # restrict that matches iff the licenses are allowed
            restrict = delegate(partial(self._apply_license_filter, master_license))
            filters.append(restrict)

        return tuple(filters)

    @klass.jit_attr_none
    def _default_licenses_manager(self):
        return Licenses(*self.source_repos_raw)

    def _apply_license_filter(self, master_licenses, pkg, mode):
        """Determine if a package's license is allowed."""
        # note we're not honoring mode; it's always match.
        # reason is that of not turning on use flags to get acceptable license
        # pairs, maybe change this down the line?

        matched_pkg_licenses = []
        for atom, licenses in self.pkg_licenses:
            if atom.match(pkg):
                matched_pkg_licenses += licenses

        raw_accepted_licenses = master_licenses + matched_pkg_licenses
        license_manager = getattr(pkg.repo, 'licenses', self._default_licenses_manager)

        for and_pair in pkg.license.dnf_solutions():
            accepted = incremental_expansion_license(
                pkg, and_pair, license_manager.groups, raw_accepted_licenses,
                msg_prefix="while checking ACCEPT_LICENSE ")
            if accepted.issuperset(and_pair):
                return True
        return False

    def _make_keywords_filter(self, default_keys, accept_keywords, incremental=False):
        """Generates a restrict that matches iff the keywords are allowed."""
        if not accept_keywords and not self.profile.keywords:
            return packages.PackageRestriction(
                "keywords", values.ContainmentMatch(frozenset(default_keys)))

        if self.unstable_arch not in default_keys:
            # stable; thus empty entries == ~arch
            def f(r, v):
                if not v:
                    return r, self.unstable_arch
                return r, v
            data = collapsed_restrict_to_data(
                ((packages.AlwaysTrue, default_keys),),
                (f(*i) for i in accept_keywords))
        else:
            if incremental:
                f = collapsed_restrict_to_data
            else:
                f = non_incremental_collapsed_restrict_to_data
            data = f(((packages.AlwaysTrue, default_keys),), accept_keywords)

        if incremental:
            raise NotImplementedError(self._incremental_apply_keywords_filter)
        else:
            f = self._apply_keywords_filter
        return delegate(partial(f, data))

    @staticmethod
    def _incremental_apply_keywords_filter(data, pkg, mode):
        # note we ignore mode; keywords aren't influenced by conditionals.
        # note also, we're not using a restriction here.  this is faster.
        allowed = data.pull_data(pkg)
        return any(True for x in pkg.keywords if x in allowed)

    def _apply_keywords_filter(self, data, pkg, mode):
        # note we ignore mode; keywords aren't influenced by conditionals.
        # note also, we're not using a restriction here.  this is faster.
        pkg_keywords = pkg.keywords
        for atom, keywords in self.profile.keywords:
            if atom.match(pkg):
                pkg_keywords += keywords
        allowed = data.pull_data(pkg)
        if '**' in allowed:
            return True
        if "*" in allowed:
            for k in pkg_keywords:
                if k[0] not in "-~":
                    return True
        if "~*" in allowed:
            for k in pkg_keywords:
                if k[0] == "~":
                    return True
        return any(True for x in pkg_keywords if x in allowed)

    @klass.jit_attr_none
    def use_expand_re(self):
        return re.compile(
            "^(?:[+-])?(%s)_(.*)$" %
            "|".join(x.lower() for x in self.profile.use_expand))

    def _split_use_expand_flags(self, use_stream):
        stream = ((self.use_expand_re.match(x), x) for x in use_stream)
        flags, ue_flags = predicate_split(bool, stream, itemgetter(0))
        return list(map(itemgetter(1), flags)), [(x[0].groups(), x[1]) for x in ue_flags]

    def get_package_use_unconfigured(self, pkg, for_metadata=True):
        """Determine use flags for a given package.

        Roughly, this should result in the following, evaluated l->r: non
        USE_EXPAND; profiles, pkg iuse, global configuration, package.use
        configuration, commandline?  stack profiles + pkg iuse; split it into
        use and use_expanded use; do global configuration + package.use
        configuration overriding of non-use_expand use if global configuration
        has a setting for use_expand.

        Args:
            pkg: package object
            for_metadata (bool): if True, we're doing use flag retrieval for
                metadata generation; otherwise, we're just requesting the raw use flags

        Returns:
            Three groups of use flags for the package in the following order:
            immutable flags, enabled flags, and disabled flags.
        """
        pre_defaults = [x[1:] for x in pkg.iuse if x[0] == '+']
        if pre_defaults:
            pre_defaults, ue_flags = self._split_use_expand_flags(pre_defaults)
            pre_defaults.extend(
                x[1] for x in ue_flags if x[0][0].upper() not in self.settings)

        attr = 'stable_' if self.stable_arch in pkg.keywords \
            and self.unstable_arch not in self.settings['ACCEPT_KEYWORDS'] else ''
        disabled = getattr(self.profile, attr + 'masked_use').pull_data(pkg)
        immutable = getattr(self, attr + 'forced_use').pull_data(pkg)

        # lock the configurable use flags to only what's in IUSE, and what's forced
        # from the profiles (things like userland_GNU and arch)
        enabled = self.enabled_use.pull_data(pkg, pre_defaults=pre_defaults)

        # support globs for USE_EXPAND vars
        use_globs = [u for u in enabled if u.endswith('*')]
        enabled_use_globs = []
        for glob in use_globs:
            for u in pkg.iuse_stripped:
                if u.startswith(glob[:-1]):
                    enabled_use_globs.append(u)
        enabled.difference_update(use_globs)
        enabled.update(enabled_use_globs)

        if for_metadata:
            preserves = pkg.iuse_stripped
            enabled.intersection_update(preserves)
            enabled.update(immutable)
            enabled.difference_update(disabled)

        return immutable, enabled, disabled

    def get_package_domain(self, pkg):
        """Get domain object with altered settings from matching package.env entries."""
        if getattr(pkg, '_domain', None) is not None:
            return pkg._domain

        files = []
        for restrict, paths in self.pkg_env:
            if restrict.match(pkg):
                files.extend(paths)
        if files:
            pkg_settings = dict(self._settings.orig.items())
            for path in files:
                PortageConfig.load_make_conf(
                    pkg_settings, path, allow_sourcing=True,
                    allow_recurse=False, incrementals=True)

            # TODO: Improve pkg domain vs main domain proxying, e.g. static
            # jitted attrs should always be generated and pulled from the main
            # domain obj; however, currently each pkg domain instance gets its
            # own copy so values collapsed on the pkg domain instance aren't
            # propagated back to the main domain leading to regen per pkg if
            # requested.
            pkg_domain = copy.copy(self)
            pkg_domain._settings = ProtectedDict(pkg_settings)
            # reset jitted attrs that can pull updated settings
            for attr in (x for x in dir(self) if x.startswith('_jit_reset_')):
                setattr(pkg_domain, attr, None)
            # store altered domain on the pkg obj to avoid recreating pkg domain
            object.__setattr__(pkg, "_domain", pkg_domain)
            return pkg_domain
        return self

    def get_package_bashrcs(self, pkg):
        for source in self.profile.bashrcs:
            yield source
        for source in self.bashrcs:
            yield source
        if not os.path.exists(self.ebuild_hook_dir):
            return
        # matching portage behavior... it's whacked.
        base = pjoin(self.ebuild_hook_dir, pkg.category)
        dirs = (
            pkg.package,
            f"{pkg.package}:{pkg.slot}",
            getattr(pkg, "P", None),
            getattr(pkg, "PF", None),
        )
        for fp in filter(None, dirs):
            fp = pjoin(base, fp)
            if os.path.exists(fp):
                yield local_source(fp)

    def _wrap_repo(self, repo, filtered=True):
        """Create a filtered, wrapped repo object for the domain."""
        wrapped_repo = self._configure_repo(repo)
        if filtered:
            wrapped_repo = self.filter_repo(wrapped_repo)
        return wrapped_repo

    def add_repo(self, path, config, configure=False, **kwargs):
        """Add an external repo to the domain."""
        path = os.path.abspath(path)

        # return repo if it's already registered
        try:
            return next(r for r in self.source_repos_raw if r.location == path)
        except StopIteration:
            pass

        # forcibly create repo_config object, otherwise cached version might be used
        try:
            repo_config = RepoConfig(path, disable_inst_caching=True)
        except OSError as e:
            raise repo_errors.InvalidRepo(str(e))

        if repo_config.cache_format is not None:
            # default to using md5 cache
            kwargs['cache'] = (md5_cache(path),)
        repo = ebuild_repo.tree(config, repo_config, **kwargs)
        self.source_repos_raw += repo

        # inject repo objects into config to dynamically register repo
        data = {}
        repo_conf_data = {
            'class': 'pkgcore.ebuild.repo_objs.RepoConfig',
            'location': path,
        }
        repo_data = {
            'inherit': ('ebuild-repo-common',),
            'repo_config': f'conf:{path}',
        }
        data[f'conf:{path}'] = basics.AutoConfigSection(repo_conf_data)
        data[path] = basics.AutoConfigSection(repo_data)
        config.update(data)

        # reset repo-related jit attrs
        for attr in (x for x in dir(self) if x.startswith('_jit_repo_')):
            setattr(self, attr, None)

        if configure:
            repo = self._wrap_repo(repo)
        return repo

    def find_repo(self, path, config, configure=False):
        """Find and add an external repo to the domain given a path."""
        path = os.path.abspath(path)
        with suppress_logging():
            while (parent := os.path.dirname(path)) != path:
                try:
                    return self.add_repo(path, config=config, configure=configure)
                except repo_errors.InvalidRepo:
                    path = parent

    def _configure_repo(self, repo):
        """Configure a raw repo."""
        if not repo.configured:
            args = []
            try:
                for x in repo.configurables:
                    if x == "domain":
                        args.append(self)
                    elif x == "settings":
                        args.append(self.settings)
                    elif x == "profile":
                        args.append(self.profile)
                    else:
                        args.append(getattr(self, x))
            except AttributeError as e:
                raise Failure(
                    f"failed configuring repo {repo!r}: "
                    f"configurable missing: {e}") from e
            repo = repo.configure(*args)
        return repo

    def filter_repo(self, repo, pkg_masks=None, pkg_unmasks=None, pkg_filters=None,
                    pkg_accept_keywords=None, pkg_keywords=None, profile=True):
        """Filter a configured repo."""
        if pkg_masks is None:
            pkg_masks = self.pkg_masks
        if pkg_unmasks is None:
            pkg_unmasks = self.pkg_unmasks
        if pkg_filters is None:
            pkg_filters = self._pkg_filters(pkg_accept_keywords, pkg_keywords)

        global_masks = [((), repo.pkg_masks)]
        if profile:
            global_masks.extend(self.profile._incremental_masks)
        masks = set()
        for neg, pos in global_masks:
            masks.difference_update(neg)
            masks.update(pos)
        masks.update(pkg_masks)
        unmasks = set()
        if profile:
            for neg, pos in self.profile._incremental_unmasks:
                unmasks.difference_update(neg)
                unmasks.update(pos)
        unmasks.update(pkg_unmasks)

        filters = generate_filter(masks, unmasks, *pkg_filters)
        return filtered.tree(repo, filters, True)

    @klass.jit_attr_named('_jit_reset_tmpdir', uncached_val=None)
    def tmpdir(self):
        """Temporary directory for the system.

        Uses PORTAGE_TMPDIR setting and falls back to using the system's TMPDIR if unset.
        """
        path = self.settings.get('PORTAGE_TMPDIR', '')
        if not os.path.exists(path):
            try:
                os.mkdir(path)
            except EnvironmentError:
                path = tempfile.gettempdir()
                logger.warning(f'nonexistent PORTAGE_TMPDIR path, defaulting to {path!r}')
        return os.path.normpath(path)

    @property
    def pm_tmpdir(self):
        """Temporary directory for the package manager."""
        return pjoin(self.tmpdir, 'portage')

    @property
    def repo_configs(self):
        """All defined repo configs."""
        return tuple(r.config for r in self.repos if hasattr(r, 'config'))

    @klass.jit_attr
    def KV(self):
        """The version of the running kernel."""
        ret, version = spawn_get_output(['uname', '-r'])
        if ret == 0:
            return version[0].strip()
        raise ValueError('unknown kernel version')

    @klass.jit_attr_named('_jit_repo_source_repos_raw', uncached_val=None)
    def source_repos_raw(self):
        """Group of package repos without filtering."""
        repos = []
        for r in self.__repos:
            try:
                repo = r.instantiate()
                if not repo.is_supported:
                    logger.warning(
                        f'skipping {r.name!r} repo: unsupported EAPI {str(repo.eapi)!r}')
                    continue
                repos.append(repo)
            except config_errors.InstantiationError as e:
                # roll back the exception chain to a meaningful error message
                exc = find_user_exception(e)
                if exc is None:
                    exc = e
                logger.warning(f'skipping {r.name!r} repo: {exc}')
        return RepositoryGroup(repos)

    @klass.jit_attr_named('_jit_repo_installed_repos_raw', uncached_val=None)
    def installed_repos_raw(self):
        """Group of installed repos without filtering."""
        repos = [r.instantiate() for r in self.__vdb]
        if self.profile.provides_repo is not None:
            repos.append(self.profile.provides_repo)
        return RepositoryGroup(repos)

    @klass.jit_attr_named('_jit_repo_repos_raw', uncached_val=None)
    def repos_raw(self):
        """Group of all repos without filtering."""
        return RepositoryGroup(
            chain(self.source_repos_raw, self.installed_repos_raw))

    @klass.jit_attr_named('_jit_repo_source_repos', uncached_val=None)
    def source_repos(self):
        """Group of configured, filtered package repos."""
        repos = []
        for repo in self.source_repos_raw:
            try:
                repos.append(self._wrap_repo(repo, filtered=True))
            except repo_errors.RepoError as e:
                logger.warning(f'skipping {repo.repo_id!r} repo: {e}')
        return RepositoryGroup(repos)

    @klass.jit_attr_named('_jit_repo_installed_repos', uncached_val=None)
    def installed_repos(self):
        """Group of configured, installed package repos."""
        repos = []
        for repo in self.installed_repos_raw:
            try:
                repos.append(self._wrap_repo(repo, filtered=False))
            except repo_errors.RepoError as e:
                logger.warning(f'skipping {repo.repo_id!r} repo: {e}')
        return RepositoryGroup(repos)

    @klass.jit_attr_named('_jit_repo_unfiltered_repos', uncached_val=None)
    def unfiltered_repos(self):
        """Group of all configured repos without filtering."""
        repos = chain(self.source_repos, self.installed_repos)
        return RepositoryGroup(
            (r.raw_repo if r.raw_repo is not None else r) for r in repos)

    @klass.jit_attr_named('_jit_repo_repos', uncached_val=None)
    def repos(self):
        """Group of all repos."""
        return RepositoryGroup(
            chain(self.source_repos, self.installed_repos))

    @klass.jit_attr_named('_jit_repo_ebuild_repos', uncached_val=None)
    def ebuild_repos(self):
        """Group of all ebuild repos bound with configuration data."""
        return RepositoryGroup(
            x for x in self.source_repos
            if isinstance(x.raw_repo, ebuild_repo.ConfiguredTree))

    @klass.jit_attr_named('_jit_repo_ebuild_repos_unfiltered', uncached_val=None)
    def ebuild_repos_unfiltered(self):
        """Group of all ebuild repos without package filtering."""
        return RepositoryGroup(
            x for x in self.unfiltered_repos
            if isinstance(x, ebuild_repo.ConfiguredTree))

    @klass.jit_attr_named('_jit_repo_ebuild_repos_raw', uncached_val=None)
    def ebuild_repos_raw(self):
        """Group of all ebuild repos without filtering."""
        return RepositoryGroup(
            x for x in self.source_repos_raw
            if isinstance(x, ebuild_repo.UnconfiguredTree))

    @klass.jit_attr_named('_jit_repo_binary_repos', uncached_val=None)
    def binary_repos(self):
        """Group of all binary repos bound with configuration data."""
        return RepositoryGroup(
            x for x in self.source_repos
            if isinstance(x.raw_repo, binary_repo.ConfiguredTree))

    @klass.jit_attr_named('_jit_repo_binary_repos_unfiltered', uncached_val=None)
    def binary_repos_unfiltered(self):
        """Group of all binary repos without package filtering."""
        return RepositoryGroup(
            x for x in self.unfiltered_repos
            if isinstance(x, binary_repo.ConfiguredTree))

    @klass.jit_attr_named('_jit_repo_binary_repos_raw', uncached_val=None)
    def binary_repos_raw(self):
        """Group of all binary repos without filtering."""
        return RepositoryGroup(
            x for x in self.source_repos_raw
            if isinstance(x, binary_repo.tree))

    # multiplexed repos
    all_repos = klass.alias_attr("repos.combined")
    all_repos_raw = klass.alias_attr("repos_raw.combined")
    all_source_repos = klass.alias_attr("source_repos.combined")
    all_source_repos_raw = klass.alias_attr("source_repos_raw.combined")
    all_installed_repos = klass.alias_attr("installed_repos.combined")
    all_installed_repos_raw = klass.alias_attr("installed_repos_raw.combined")
    all_unfiltered_repos = klass.alias_attr("unfiltered_repos.combined")
    all_ebuild_repos = klass.alias_attr("ebuild_repos.combined")
    all_ebuild_repos_unfiltered = klass.alias_attr("ebuild_repos_unfiltered.combined")
    all_ebuild_repos_raw = klass.alias_attr("ebuild_repos_raw.combined")
    all_binary_repos = klass.alias_attr("binary_repos.combined")
    all_binary_repos_unfiltered = klass.alias_attr("binary_repos_unfiltered.combined")
    all_binary_repos_raw = klass.alias_attr("binary_repos_raw.combined")



import operator

from pkgcore.cache import base, bulk, errors
from snakeoil.chksum import LazilyHashedPath
from snakeoil.test import TestCase


def _mk_chf_obj(**kwargs):
    kwargs.setdefault('mtime', 100)
    return LazilyHashedPath('/nonexistent/path', **kwargs)
_chf_obj = _mk_chf_obj()


class DictCache(base):
    """Minimal dict-backed cache for testing."""

    cleanse_keys = True
    autocommits = True
    __has_working_commit__ = False

    def __init__(self, *args, **kwargs):
        base.__init__(self, *args, **kwargs)
        self._data = {}

    def _getitem(self, cpv):
        return self._data[cpv]

    def __getitem__(self, cpv):
        # Protected dict's come back by default, but are a minor
        # pita to deal with for this code- thus we convert back.
        # Additionally, we drop any chksum info in the process.
        d = dict(base.__getitem__(self, cpv).items())
        d.pop(f'_{self.chf_type}_', None)
        return d

    def __setitem__(self, cpv, data):
        data['_chf_'] = _chf_obj
        return base.__setitem__(self, cpv, data)

    def _setitem(self, cpv, values):
        self._data[cpv] = values

    def _delitem(self, cpv):
        del self._data[cpv]

    def __contains__(self, cpv):
        return cpv in self._data

    def keys(self):
        return iter(self._data.keys())


class DictCacheBulk(bulk):

    cleanse_keys = True
    __has_working_commit__ = True

    def __init__(self, *args, **kwargs):
        bulk.__init__(self, *args, **kwargs)
        self._data = {}
        self._write_count = 0

    def _read_data(self):
        return self._data.copy()

    def _write_data(self):
        self._data = self.data.copy()
        self._write_count += 1

    def __getitem__(self, cpv):
        d = bulk.__getitem__(self, cpv)
        d.pop(f'_{self.chf_type}_', None)
        return d

    def __setitem__(self, cpv, data):
        data['_chf_'] = _chf_obj
        return bulk.__setitem__(self, cpv, data)

    def keys(self):
        return iter(self._data.keys())


class BaseTest(TestCase):

    cache_keys = ("foo", "_eclasses_")

    def get_db(self, readonly=False):
        return DictCache(auxdbkeys=self.cache_keys,
            readonly=readonly)

    def test_basics(self):
        self.cache = self.get_db()
        self.cache['spork'] = {'foo':'bar'}
        assert {'foo': 'bar'} == self.cache['spork']
        self.assertRaises(KeyError, operator.getitem, self.cache, 'notaspork')

        self.cache['spork'] = {'foo': 42}
        self.cache['foon'] = {'foo': 42}
        assert {'foo': 42} == self.cache['spork']
        assert {'foo': 42} == self.cache['foon']

        assert ['foon', 'spork'] == sorted(self.cache.keys())
        assert [('foon', {'foo': 42}), ('spork', {'foo': 42})] == sorted(self.cache.items())
        del self.cache['foon']
        self.assertRaises(KeyError, operator.getitem, self.cache, 'foon')

        assert 'spork' in self.cache
        assert 'foon' not in self.cache

        self.cache['empty'] = {'foo': ''}
        assert {} == self.cache['empty']

    def test_eclasses(self):
        self.cache = self.get_db()
        self.cache['spork'] = {'foo':'bar'}
        self.cache['spork'] = {'_eclasses_': {'spork': _chf_obj,
                                              'foon': _chf_obj}}
        self.assertRaises(errors.CacheCorruption,
                          operator.getitem, self.cache, 'spork')

        self.cache['spork'] = {'_eclasses_': {'spork': _mk_chf_obj(mtime=1),
                                              'foon': _mk_chf_obj(mtime=2)}}
        assert self.cache._data['spork']['_eclasses_'] in ['spork\t1\tfoon\t2', 'foon\t2\tspork\t1']
        assert (
            sorted([('foon', (('mtime', 2),)), ('spork', (('mtime', 1),))]) ==
            sorted(self.cache['spork']['_eclasses_']))

    def test_readonly(self):
        self.cache = self.get_db()
        self.cache['spork'] = {'foo':'bar'}
        cache = self.get_db(True)
        cache._data = self.cache._data
        self.assertRaises(errors.ReadOnly,
                          operator.delitem, cache, 'spork')
        self.assertRaises(errors.ReadOnly,
                          operator.setitem, cache, 'spork', {'foo': 42})
        assert {'foo': 'bar'} == cache['spork']

    def test_clear(self):
        cache = self.get_db()
        cache['spork'] = {'foo':'bar'}
        assert {'foo':'bar'} == cache['spork']
        assert list(cache) == ['spork']
        cache['dork'] = {'foo':'bar2'}
        cache['dork2'] = {'foo':'bar2'}
        assert list(sorted(cache)) == sorted(['dork', 'dork2', 'spork'])
        cache.clear()
        assert list(cache) == []

    def test_sync_rate(self):
        db = self.get_db()
        tracker = []

        def commit(tracker=tracker, raw_commit=db.commit):
            tracker.append(True)
            if db.__has_working_commit__:
                raw_commit()

        db.commit = commit
        db.autocommits = False

        db.set_sync_rate(2)
        db["dar"] = {"foo":"blah"}
        assert not tracker
        db["dar2"] = {"foo":"blah"}
        assert len(tracker) == 1
        assert sorted(db) == ["dar", "dar2"]
        db["dar3"] = {"foo":"blah"}
        assert len(tracker) == 1

        # finally ensure sync_rate(1) behaves
        db.set_sync_rate(1)
        # ensure it doesn't flush just for fiddling w/ sync rate
        assert len(tracker) == 1
        db["dar4"] = {"foo":"blah"}
        assert len(tracker) == 2
        db["dar5"] = {"foo":"blah"}
        assert len(tracker) == 3


class TestBulk(BaseTest):

    def get_db(self, readonly=False):
        return DictCacheBulk(auxdbkeys=self.cache_keys,
            readonly=readonly)

    def test_filtering(self):
        db = self.get_db()
        # write a key outside of known keys
        db["dar"] = {"foo2":"dar"}
        assert list(db["dar"].items()) == []



"""
EBuild Daemon (ebd), main high level interface to ebuild execution env.

Wraps :obj:`pkgcore.ebuild.processor` functionality into a higher level
api, for example per phase methods.
"""

__all__ = (
    "ebd", "setup_mixin", "install_op", "uninstall_op", "replace_op",
    "buildable", "binpkg_localize")

import errno
import os
import re
import shutil
import sys
import time
from collections import defaultdict
from functools import partial
from itertools import chain
from tempfile import TemporaryFile

from snakeoil import data_source, klass
from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.contexts import chdir
from snakeoil.currying import post_curry, pretty_docs
from snakeoil.fileutils import touch
from snakeoil.osutils import ensure_dirs, listdir_files, normpath, pjoin
from snakeoil.process.spawn import is_sandbox_capable, is_userpriv_capable, spawn, spawn_bash
from snakeoil.sequences import iflatten_instance, iter_stable_unique

from .. import const
from ..log import logger
from ..operations import format, observer
from ..os_data import portage_gid, portage_uid, xargs
from ..package.mutated import MutatedPkg
from . import ebd_ipc, ebuild_built, errors
from .processor import (ProcessorError, chuck_UnhandledCommand, expected_ebuild_env,
                        inherit_handler, release_ebuild_processor, request_ebuild_processor)


class ebd:

    def __init__(self, pkg, initial_env=None, env_data_source=None,
                 observer=None, clean=True, tmp_offset=None):
        """
        :param pkg:
            :class:`pkgcore.ebuild.ebuild_src.package`
            instance this env is being setup for
        :param initial_env: initial environment to use for this ebuild
        :param env_data_source: a :obj:`snakeoil.data_source.base` instance
            to restore the environment from- used for restoring the
            state of an ebuild processing, whether for unmerging, or
            walking phases during building
        """
        self.pkg = pkg
        self.eapi = pkg.eapi

        if not hasattr(self, "observer"):
            self.observer = observer
        if not self.eapi.is_supported:
            raise TypeError(f"package {pkg} uses unsupported EAPI: {str(self.eapi)!r}")

        if initial_env is not None:
            # copy.
            self.env = dict(initial_env)
            for x in ("USE", "ACCEPT_LICENSE"):
                self.env.pop(x, None)
        else:
            self.env = {}

        # Drop all USE_EXPAND variables from the exported environment.
        for u in self.domain.profile.use_expand:
            self.env.pop(u, None)

        # Only export USE_EXPAND variables for the package's enabled USE flags.
        d = defaultdict(list)
        for u in pkg.use:
            m = self.domain.use_expand_re.match(u)
            if m:
                use_expand, value = m.groups()
                d[use_expand.upper()].append(value)
        for k, v in d.items():
            self.env[k] = ' '.join(sorted(v))

        self.bashrc = self.env.pop("bashrc", ())
        self.features = set(x.lower() for x in self.domain.features)
        self.env["FEATURES"] = ' '.join(sorted(self.features))
        self.set_path_vars(self.env, self.pkg, self.domain)

        # internally implemented EAPI specific functions to skip when exporting env
        self.env["PKGCORE_EAPI_FUNCS"] = ' '.join(self.eapi.bash_funcs)

        self.env_data_source = env_data_source
        if (env_data_source is not None and
                not isinstance(env_data_source, data_source.base)):
            raise TypeError(
                "env_data_source must be None, or a pkgcore.data_source.base "
                f"derivative: {env_data_source.__class__}: {env_data_source}")

        iuse_effective_regex = f"^({'|'.join(re.escape(x) for x in pkg.iuse_effective)})$"
        self.env["PKGCORE_IUSE_EFFECTIVE"] = iuse_effective_regex.replace("\\.\\*", ".*")

        expected_ebuild_env(pkg, self.env, env_source_override=self.env_data_source)

        self.env["PKGCORE_FINALIZED_RESTRICT"] = ' '.join(str(x) for x in pkg.restrict)

        self.restrict = pkg.restrict

        for x in ("sandbox", "userpriv"):
            setattr(self, x, self.feat_or_bool(x) and not (x in self.restrict))
        if self.userpriv and os.getuid() != 0:
            self.userpriv = False

        if "PORT_LOGDIR" in self.env:
            self.logging = pjoin(
                self.env["PORT_LOGDIR"],
                "%s:%s:%s.log" % (
                    pkg.cpvstr, self.__class__.__name__,
                    time.strftime("%Y%m%d-%H%M%S", time.localtime())))
            del self.env["PORT_LOGDIR"]
        else:
            self.logging = False

        self.env["PKGCORE_PKG_REPO"] = pkg.source_repository
        self.env["XARGS"] = xargs

        # wipe variables listed in ENV_UNSET for supporting EAPIs
        if self.eapi.options.has_env_unset:
            for x in self.env.pop('ENV_UNSET', ()):
                self.env.pop(x, None)

        # wipe any remaining internal settings from the exported env
        wipes = [k for k, v in self.env.items()
                 if not isinstance(v, str)]
        for k in wipes:
            del self.env[k]

        self._set_op_vars(tmp_offset)
        self.clean_at_start = clean
        self.clean_needed = False

        # various IPC command support
        self._ipc_helpers = {
            # bash helpers
            'doins': ebd_ipc.Doins(self),
            'dodoc': ebd_ipc.Dodoc(self),
            'dohtml': ebd_ipc.Dohtml(self),
            'doinfo': ebd_ipc.Doinfo(self),
            'dodir': ebd_ipc.Dodir(self),
            'doexe': ebd_ipc.Doexe(self),
            'dobin': ebd_ipc.Dobin(self),
            'dosbin': ebd_ipc.Dosbin(self),
            'dolib': ebd_ipc.Dolib(self),
            'dolib.so': ebd_ipc.Dolib_so(self),
            'dolib.a': ebd_ipc.Dolib_a(self),
            'doman': ebd_ipc.Doman(self),
            'domo': ebd_ipc.Domo(self),
            'dosym': ebd_ipc.Dosym(self),
            'dohard': ebd_ipc.Dohard(self),
            'keepdir': ebd_ipc.Keepdir(self),

            # bash functions
            'has_version': ebd_ipc.Has_Version(self),
            'best_version': ebd_ipc.Best_Version(self),
            'unpack': ebd_ipc.Unpack(self),
            'eapply': ebd_ipc.Eapply(self),
            'eapply_user': ebd_ipc.Eapply_User(self),
            'docompress': ebd_ipc.Docompress(self),
            'dostrip': ebd_ipc.Dostrip(self),

            # internals
            'filter_env': ebd_ipc.FilterEnv(self),
        }

    def start(self):
        if self.clean_at_start:
            self.clean_needed = True
            if not self.cleanup():
                return False
        self.setup_workdir()
        self._setup_env_data_source()
        self.clean_needed = True
        return True

    @staticmethod
    def set_path_vars(env, pkg, domain):
        # XXX: note this is just EAPI 3 and EAPI 7 compatibility; not full prefix, soon..
        trailing_slash = pkg.eapi.options.trailing_slash
        env['ROOT'] = domain.root.rstrip(os.sep) + trailing_slash
        env['PKGCORE_PREFIX_SUPPORT'] = 'false'
        if pkg.eapi.options.prefix_capable:
            env['EPREFIX'] = domain.prefix.rstrip(os.sep)
            env['EROOT'] = (
                pjoin(env['ROOT'].rstrip(trailing_slash), env['EPREFIX'])
                + trailing_slash)
            env['PKGCORE_PREFIX_SUPPORT'] = 'true'

        if pkg.eapi.options.has_sysroot:
            env['SYSROOT'] = env['ROOT']
            env['ESYSROOT'] = pjoin(env['SYSROOT'], env['EPREFIX'])
            env['BROOT'] = env['EPREFIX']

    def _set_op_vars(self, tmp_offset):
        # don't fool with this, without fooling with setup.
        self.tmpdir = self.domain.pm_tmpdir
        if tmp_offset:
            self.tmpdir = pjoin(self.tmpdir, tmp_offset.strip(os.sep))

        self.builddir = pjoin(self.tmpdir, self.env["CATEGORY"], self.env["PF"])
        for x, y in (("T", "temp"),
                     ("WORKDIR", "work"),
                     ("D", "image"),
                     ("HOME", "homedir"),
                     ("PKGCORE_EMPTYDIR", "empty")):
            self.env[x] = normpath(pjoin(self.builddir, y))
        self.env["D"] += self.eapi.options.trailing_slash
        self.env["PORTAGE_LOGFILE"] = normpath(pjoin(self.env["T"], "build.log"))

        # XXX: Note that this is just EAPI 3 support, not yet prefix
        # full awareness.
        if self.pkg.eapi.options.prefix_capable:
            self.env["ED"] = normpath(
                pjoin(self.env["D"].rstrip(os.sep), self.env["EPREFIX"])) \
                    + self.eapi.options.trailing_slash

        # temporary install dir correct for all EAPIs
        self.ED = self.env.get('ED', self.env['D'])

    def get_env_source(self):
        with open(pjoin(self.env["T"], "environment"), "rb") as f:
            return data_source.bytes_data_source(f.read())

    def _setup_env_data_source(self):
        if not ensure_dirs(self.env["T"], mode=0o770, gid=portage_gid, minimal=True):
            raise format.FailedDirectory(
                self.env['T'],
                "%s doesn't fulfill minimum mode %o and gid %i" % (
                    self.env['T'], 0o770, portage_gid))

        if self.env_data_source is not None:
            fp = pjoin(self.env["T"], "environment")
            # load data first (might be a local_source), *then* write
            # if it's a src_ebuild being installed, trying to do two steps
            # stomps the local_sources data.
            data = self.env_data_source.bytes_fileobj().read()
            with open(fp, "wb") as f:
                f.write(data)
            del data

    def _set_per_phase_env(self, phase, env):
        self._setup_merge_type(phase, env)

        # add phase specific helper paths to PATH if they exist
        ebuild_phase = self.eapi.phases.get(phase, '')
        if ebuild_phase in self.eapi.helpers:
            path = chain.from_iterable((
                const.PATH_FORCED_PREPEND,
                self.pkg.eapi.helpers.get('global', ()),
                self.eapi.helpers[ebuild_phase],
                os.environ.get('PATH', '').split(os.pathsep),
            ))
            env['PATH'] = os.pathsep.join(path)

    def _setup_merge_type(self, phase, env):
        # only allowed in pkg_ phases.

        if (not self.eapi.phases.get(phase, "").startswith("pkg_") and
                not phase == 'setup-binpkg'):
            return

        # note all pkgs have this attribute
        is_source = getattr(self.pkg, '_is_from_source', True)

        if self.eapi.options.has_merge_type:
            env["MERGE_TYPE"] = (is_source and "source") or "binary"
        else:
            # we still must export this, just via the portage var name w/
            # different values.  if we didn't, spec or not, kernel binpkg
            # merging would be broke.
            env["EMERGE_FROM"] = (is_source and "ebuild") or "binary"

    def setup_logging(self):
        if self.logging and not ensure_dirs(os.path.dirname(self.logging),
                                            mode=0o2770, gid=portage_gid):
            raise format.FailedDirectory(
                os.path.dirname(self.logging),
                "PORT_LOGDIR, desired mode 02770 and gid %i" % portage_gid)

    def setup_workdir(self):
        # ensure dirs.
        for k in ("HOME", "T", "WORKDIR", "D"):
            if not ensure_dirs(self.env[k], mode=0o4770, gid=portage_gid, minimal=True):
                raise format.FailedDirectory(
                    self.env[k],
                    "%s doesn't fulfill minimum mode %o and gid %i" % (k, 0o770, portage_gid))
            # XXX hack, just 'til pkgcore controls these directories
            if (os.stat(self.env[k]).st_mode & 0o2000):
                logger.warning(f"{self.env[k]} ( {k} ) is setgid")

    def _generic_phase(self, phase, userpriv, sandbox, extra_handlers={},
                       failure_allowed=False, suppress_bashrc=False):
        """
        :param phase: phase to execute
        :param userpriv: will we drop to
            :obj:`pkgcore.os_data.portage_uid` and
            :obj:`pkgcore.os_data.portage_gid` access for this phase?
        :param sandbox: should this phase be sandboxed?
        """
        if phase not in self.pkg.mandatory_phases:
            # TODO(ferringb): Note the preinst hack; this will be removed once dyn_pkg_preinst
            # is dead in full (currently it has a selinux labelling and suidctl ran from there)
            if phase != 'preinst':
                return True
            if 'selinux' not in self.features and 'suidctl' not in self.features:
                return True

        shutil.rmtree(self.env["PKGCORE_EMPTYDIR"], ignore_errors=True)
        os.mkdir(self.env["PKGCORE_EMPTYDIR"])

        userpriv = self.userpriv and userpriv
        sandbox = self.sandbox and sandbox
        self._set_per_phase_env(phase, self.env)
        extra_handlers = extra_handlers.copy()
        extra_handlers.update(self._ipc_helpers)
        if not suppress_bashrc:
            extra_handlers.setdefault("request_bashrcs", self._request_bashrcs)
        return run_generic_phase(
            self.pkg, phase, self.env, userpriv, sandbox,
            extra_handlers=extra_handlers, failure_allowed=failure_allowed,
            logging=self.logging)

    def _request_bashrcs(self, ebd):
        for source in self.domain.get_package_bashrcs(self.pkg):
            if source.path is not None:
                ebd.write(f"path\n{source.path}")
            elif source.get_data is not None:
                raise NotImplementedError
            else:
                chuck_UnhandledCommand(
                    ebd, "bashrc request: unable to process bashrc "
                    f"due to source '{source}' due to lacking usable get_*")
            if not ebd.expect("next"):
                chuck_UnhandledCommand(
                    ebd, "bashrc transfer, didn't receive 'next' response.  "
                    "failure?")
        ebd.write("end_request")

    def set_is_replacing(self, *pkgs):
        if self.eapi.options.exports_replacing:
            self.env['REPLACING_VERSIONS'] = " ".join(pkg.PVR for pkg in pkgs)

    def set_is_being_replaced_by(self, pkg=None):
        if self.eapi.options.exports_replacing and pkg is not None:
            self.env['REPLACED_BY_VERSION'] = pkg.PVR

    def cleanup(self, disable_observer=False, force=False):
        if not force:
            if not self.clean_needed:
                return True
        if not os.path.exists(self.builddir):
            return True
        if disable_observer:
            return self.do_cleanup(disable_observer=disable_observer)
        return self.do_cleanup()

    @observer.decorate_build_method("cleanup")
    def do_cleanup(self):
        try:
            shutil.rmtree(self.builddir)
            # try to wipe the cat dir; if not empty, ignore it
            try:
                os.rmdir(os.path.dirname(self.builddir))
            except EnvironmentError as e:
                # POSIX specifies either ENOTEMPTY or EEXIST for non-empty dir
                # in particular, Solaris uses EEXIST in that case.
                # https://github.com/pkgcore/pkgcore/pull/181
                if e.errno not in (errno.ENOTEMPTY, errno.EEXIST):
                    raise
        except EnvironmentError as e:
            raise format.GenericBuildError(
                f"clean: Caught exception while cleansing: {e}") from e
        return True

    def feat_or_bool(self, name, extra_env=None):
        if name in self.env:
            v = bool(self.env[name])
            del self.env[name]
            name = name.lower()
            if v:
                self.features.add(name)
            else:
                if name in self.features:
                    self.features.remove(name)
        elif extra_env is not None and name in extra_env:
            v = bool(extra_env[name])
            if v:
                self.features.add(name.lower())
            else:
                self.features.remove(name.lower())
        else:
            v = name.lower() in self.features
        return v

    def __stage_step_callback__(self, stage):
        try:
            touch(pjoin(self.builddir, f'.{stage}'))
        except EnvironmentError:
            # we really don't care...
            pass

    def _reload_state(self):
        try:
            self.__set_stage_state__(
                [x[1:] for x in listdir_files(self.builddir) if x.startswith(".")])
        except EnvironmentError as e:
            if e.errno not in (errno.ENOTDIR, errno.ENOENT):
                raise


class setup_mixin:

    setup_is_for_src = True

    def setup(self, setup_phase_override=None):
        self.setup_logging()

        additional_commands = {}
        phase_name = "setup-binpkg"
        if self.setup_is_for_src:
            phase_name = "setup"
        if setup_phase_override is not None:
            phase_name = setup_phase_override

        if self.setup_is_for_src:
            additional_commands["request_inherit"] = partial(inherit_handler, self.eclass_cache)

        return self._generic_phase(
            phase_name, False, True, extra_handlers=additional_commands)


def run_generic_phase(pkg, phase, env, userpriv, sandbox, fd_pipes=None,
                      extra_handlers=None, failure_allowed=False, logging=None, **kwargs):
    """
    :param phase: phase to execute
    :param env: environment mapping for the phase
    :param userpriv: will we drop to
        :obj:`pkgcore.os_data.portage_uid` and
        :obj:`pkgcore.os_data.portage_gid` access for this phase?
    :param sandbox: should this phase be sandboxed?
    :param fd_pipes: use custom file descriptors for ebd instance
    :type fd_pipes: mapping between file descriptors
    :param extra_handlers: extra command handlers
    :type extra_handlers: mapping from string to callable
    :param failure_allowed: allow failure without raising error
    :type failure_allowed: boolean
    :param logging: None or a filepath to log output to
    :return: True when the phase has finished execution
    """

    userpriv = userpriv and is_userpriv_capable()
    sandbox = sandbox and is_sandbox_capable()
    tmpdir = kwargs.get('tmpdir', env.get('T', None))

    if env is None:
        env = expected_ebuild_env(pkg)

    ebd = request_ebuild_processor(userpriv=userpriv, sandbox=sandbox, fd_pipes=fd_pipes)
    # this is a bit of a hack; used until ebd accepts observers that handle
    # the output redirection on its own.  Primary relevance is when
    # stdout/stderr are pointed at a file; we leave buffering on, just
    # force the flush for synchronization.
    sys.stdout.flush()
    sys.stderr.flush()
    try:
        if not ebd.run_phase(phase, env, tmpdir=tmpdir, sandbox=sandbox,
                             logging=logging, additional_commands=extra_handlers):
            if not failure_allowed:
                raise format.GenericBuildError(
                    phase + ": Failed building (False/0 return from handler)")
                logger.warning(f"executing phase {phase}: execution failed, ignoring")
    except Exception as e:
        if isinstance(e, ebd_ipc.IpcError):
            # notify bash side of IPC error
            ebd.write(e.ret)
            if isinstance(e, ebd_ipc.IpcInternalError):
                # show main exception cause for internal IPC errors
                ebd.shutdown_processor(force=True)
                raise e.__cause__
        try:
            ebd.shutdown_processor()
        except ProcessorError as pe:
            # catch die errors during shutdown
            e = pe
        release_ebuild_processor(ebd)
        if isinstance(e, ProcessorError):
            # force verbose die output
            e._verbosity = 1
            raise e
        elif isinstance(e, IGNORED_EXCEPTIONS + (format.GenericBuildError,)):
            raise
        raise format.GenericBuildError(
            f"Executing phase {phase}: Caught exception: {e}") from e

    release_ebuild_processor(ebd)
    return True


class install_op(ebd, format.install):
    """Phase operations and steps for install execution."""

    def __init__(self, domain, pkg, observer):
        format.install.__init__(self, domain, pkg, observer)
        ebd.__init__(
            self, pkg, observer=observer, initial_env=self.domain.settings,
            env_data_source=pkg.environment, clean=False)

    preinst = pretty_docs(
        observer.decorate_build_method("preinst")(
            post_curry(ebd._generic_phase, "preinst", False, False)),
        "run the postinst phase")
    postinst = pretty_docs(
        observer.decorate_build_method("postinst")(
            post_curry(ebd._generic_phase, "postinst", False, False)),
        "run the postinst phase")

    def add_triggers(self, domain_op, engine):
        self.new_pkg.add_format_triggers(domain_op, self, engine)


class uninstall_op(ebd, format.uninstall):
    """Phase operations and steps for uninstall execution."""

    def __init__(self, domain, pkg, observer):
        format.uninstall.__init__(self, domain, pkg, observer)
        ebd.__init__(
            self, pkg, observer=observer, initial_env=self.domain.settings,
            env_data_source=pkg.environment, clean=False,
            tmp_offset="unmerge")

    prerm = pretty_docs(
        observer.decorate_build_method("prerm")(
            post_curry(ebd._generic_phase, "prerm", False, False)),
        "run the prerm phase")
    postrm = pretty_docs(
        observer.decorate_build_method("postrm")(
            post_curry(
                ebd._generic_phase, "postrm", False, False,
                failure_allowed=True)),
        "run the postrm phase")

    def add_triggers(self, domain_op, engine):
        self.old_pkg.add_format_triggers(domain_op, self, engine)

    def finish(self):
        self.cleanup()
        return format.uninstall.finish(self)


class replace_op(format.replace):
    """Phase operations and steps for replace execution."""

    install_kls = staticmethod(install_op)
    uninstall_kls = staticmethod(uninstall_op)

    def __init__(self, domain, old_pkg, new_pkg, observer):
        super().__init__(domain, old_pkg, new_pkg, observer)
        self.install_op = install_op(domain, new_pkg, observer)
        self.install_op.set_is_replacing(old_pkg)
        self.uninstall_op = uninstall_op(domain, old_pkg, observer)
        self.uninstall_op.set_is_being_replaced_by(new_pkg)

    def start(self):
        self.install_op.start()
        self.uninstall_op.start()
        return True

    prerm = klass.alias_method("uninstall_op.prerm")
    postrm = klass.alias_method("uninstall_op.postrm")
    preinst = klass.alias_method("install_op.preinst")
    postinst = klass.alias_method("install_op.postinst")

    def finalize(self):
        ret = self.uninstall_op.finish()
        ret2 = self.install_op.finish()
        return (ret and ret2)

    def add_triggers(self, domain_op, engine):
        self.uninstall_op.add_triggers(domain_op, engine)
        self.install_op.add_triggers(domain_op, engine)


class buildable(ebd, setup_mixin, format.build):
    """Generic build operation."""

    # XXX this is unclean- should be handing in strictly what is build
    # env, rather then dumping domain settings as env.
    def __init__(self, domain, pkg, verified_files, eclass_cache,
                 observer=None, force_test=False, **kwargs):
        """
        :param pkg: :obj:`pkgcore.ebuild.ebuild_src.package` instance we'll be
            building
        :param eclass_cache: the :class:`pkgcore.ebuild.eclass_cache`
            we'll be using
        :param verified_files: mapping of fetchables mapped to their disk location
        """
        self._built_class = ebuild_built.fresh_built_package
        format.build.__init__(self, domain, pkg, verified_files, observer)
        domain_settings = self.domain.settings
        ebd.__init__(self, pkg, initial_env=domain_settings, **kwargs)

        self.env["FILESDIR"] = pjoin(os.path.dirname(pkg.ebuild.path), "files")
        self.eclass_cache = eclass_cache

        self.run_test = force_test or self.feat_or_bool("test", domain_settings)
        self.allow_failed_test = self.feat_or_bool("test-fail-continue", domain_settings)
        if "test" in self.restrict:
            self.run_test = False
        elif not force_test and "test" not in pkg.use:
            if self.run_test:
                logger.warning(f"disabling test for {pkg} due to test use flag being disabled")
            self.run_test = False

        # XXX minor hack
        path = self.env["PATH"].split(os.pathsep)

        for s, default in (("DISTCC", ".distcc"), ("CCACHE", "ccache")):
            b = (self.feat_or_bool(s, domain_settings) and
                 s not in self.restrict)
            setattr(self, s.lower(), b)
            if b:
                # looks weird I realize, but
                # pjoin("/foor/bar", "/barr/foo") == "/barr/foo"
                # and pjoin("/foo/bar", ".asdf") == "/foo/bar/.asdf"
                self.env.setdefault(s + "_DIR", pjoin(self.domain.tmpdir, default))
                # gentoo bug 355283
                libdir = self.env.get("ABI")
                if libdir is not None:
                    libdir = self.env.get(f"LIBDIR_{libdir}")
                    if libdir is not None:
                        libdir = self.env.get(libdir)
                if libdir is None:
                    libdir = "lib"
                path.insert(0, f"/usr/{libdir}/{s.lower()}/bin")
            else:
                for y in ("_PATH", "_DIR"):
                    if s + y in self.env:
                        del self.env[s+y]
        self.env["PATH"] = os.pathsep.join(path)

        # ordering must match appearance order in SRC_URI per PMS
        self.env["A"] = ' '.join(iter_stable_unique(pkg.distfiles))

        if self.eapi.options.has_AA:
            pkg = self.pkg
            while hasattr(pkg, '_raw_pkg'):
                pkg = getattr(pkg, '_raw_pkg')
            self.env["AA"] = ' '.join(set(iflatten_instance(pkg.distfiles)))

        if self.eapi.options.has_KV:
            self.env["KV"] = domain.KV

        if self.eapi.options.has_merge_type:
            self.env["MERGE_TYPE"] = "source"

        if self.eapi.options.has_portdir:
            self.env["PORTDIR"] = pkg.repo.location
            self.env["ECLASSDIR"] = eclass_cache.eclassdir

        if self.setup_is_for_src:
            # TODO: PORTAGE_ACTUAL_DISTDIR usage by VCS eclasses needs to be
            # dropped, but it's currently required for repo reuse.
            self.env['PORTAGE_ACTUAL_DISTDIR'] = domain.distdir
            self.env['DISTDIR'] = normpath(pjoin(self.builddir, 'distdir'))
            for k in ('PORTAGE_ACTUAL_DISTDIR', 'DISTDIR'):
                self.env[k] = os.path.realpath(self.env[k]).rstrip(os.sep) + os.sep

    def _setup_distfiles(self):
        # fetch distfiles
        if not self.verified_files:
            ops = self.domain.pkg_operations(self.pkg, observer=self.observer)
            if ops.fetch():
                # this break encapsulation and should be refactored.  Trace
                # f35f2 and 6561eac for where this was refactored.
                self.verified_files = ops.verified_files

        # symlink them into builddir
        if self.verified_files:
            try:
                if os.path.exists(self.env["DISTDIR"]):
                    if (os.path.isdir(self.env["DISTDIR"]) and
                            not os.path.islink(self.env["DISTDIR"])):
                        shutil.rmtree(self.env["DISTDIR"])
                    else:
                        os.unlink(self.env["DISTDIR"])

            except EnvironmentError as e:
                raise format.FailedDirectory(
                    self.env["DISTDIR"],
                    f"failed removing existing file/dir/link: {e}") from e

            if not ensure_dirs(self.env["DISTDIR"], mode=0o770, gid=portage_gid):
                raise format.FailedDirectory(
                    self.env["DISTDIR"],
                    "failed creating distdir symlink directory")

            try:
                for src, dest in [
                        (k, pjoin(self.env["DISTDIR"], v.filename))
                        for (k, v) in self.verified_files.items()]:
                    os.symlink(src, dest)

            except EnvironmentError as e:
                raise format.GenericBuildError(
                    f"Failed symlinking in distfiles for src {src} -> {dest}: {e}") from e

    @observer.decorate_build_method("setup")
    def setup(self):
        """Execute the setup phase, mapping out to pkg_setup in the ebuild.

        Necessarily dirs are created as required, and build env is
        initialized at this point.
        """
        if self.distcc:
            for p in ("", "/lock", "/state"):
                if not ensure_dirs(pjoin(self.env["DISTCC_DIR"], p),
                                   mode=0o2775, gid=portage_gid):
                    raise format.FailedDirectory(
                        pjoin(self.env["DISTCC_DIR"], p),
                        "failed creating needed distcc directory")
        if self.ccache:
            # yuck.
            st = None
            try:
                st = os.stat(self.env["CCACHE_DIR"])
            except OSError as e:
                st = None
                if not ensure_dirs(self.env["CCACHE_DIR"], mode=0o2775,
                                   gid=portage_gid):
                    raise format.FailedDirectory(
                        self.env["CCACHE_DIR"],
                        "failed creation of ccache dir") from e

                # XXX this is more then mildly stupid.
                st = os.stat(self.env["CCACHE_DIR"])
            try:
                if st.st_gid != portage_gid or (st.st_mode & 0o2775) != 0o2775:
                    try:
                        cwd = os.getcwd()
                    except OSError:
                        cwd = "/"
                    with chdir(cwd):
                        # crap.
                        os.chmod(self.env["CCACHE_DIR"], 0o2775)
                        os.chown(self.env["CCACHE_DIR"], -1, portage_gid)
                        if 0 != spawn(
                                ["chgrp", "-R", str(portage_gid), self.env["CCACHE_DIR"]]):
                            raise format.FailedDirectory(
                                self.env["CCACHE_DIR"],
                                "failed changing ownership for CCACHE_DIR")
                        if 0 != spawn_bash(
                                "find '%s' -type d -print0 | %s --null chmod 02775"
                                % (self.env["CCACHE_DIR"], xargs)):
                            raise format.FailedDirectory(
                                self.env["CCACHE_DIR"],
                                "failed correcting perms for CCACHE_DIR")

                        if 0 != spawn_bash(
                                "find '%s' -type f -print0 | %s --null chmod 0775"
                                % (self.env["CCACHE_DIR"], xargs)):
                            raise format.FailedDirectory(
                                self.env["CCACHE_DIR"],
                                "failed correcting perms for CCACHE_DIR")
            except OSError as e:
                raise format.FailedDirectory(
                    self.env["CCACHE_DIR"],
                    "failed ensuring perms/group owner for CCACHE_DIR") from e

        return setup_mixin.setup(self)

    def configure(self):
        """Execute the configure phase.

        Does nothing if the pkg's EAPI is less than 2 (that spec lacks a
        separated configure phase).
        """
        if "configure" in self.eapi.phases:
            return self._generic_phase("configure", True, True)
        return True

    def prepare(self):
        """Execute a source preparation phase.

        does nothing if the pkg's EAPI is less than 2
        """
        ret = True
        if "prepare" in self.eapi.phases:
            ret = self._generic_phase("prepare", True, True)
            if (self.eapi.options.user_patches and
                    not os.path.exists(pjoin(self.env['T'], '.user_patches_applied'))):
                self.observer.error(
                    'eapply_user (or default) must be called in src_prepare()')
                raise format.GenericBuildError('missing eapply_user call')
        return ret

    def nofetch(self):
        """Execute the nofetch phase.

        We need the same prerequisites as setup, so reuse that.
        """
        ensure_dirs(self.env["T"], mode=0o770, gid=portage_gid, minimal=True)
        return setup_mixin.setup(self, "nofetch")

    def unpack(self):
        """Execute the unpack phase."""
        if self.setup_is_for_src:
            self._setup_distfiles()
        if self.userpriv:
            try:
                os.chown(self.env["WORKDIR"], portage_uid, -1)
            except OSError as e:
                raise format.GenericBuildError(
                    "failed forcing %i uid for WORKDIR: %s" %
                    (portage_uid, e)) from e
        return self._generic_phase("unpack", True, True)

    compile = pretty_docs(
        observer.decorate_build_method("compile")(
            post_curry(ebd._generic_phase, "compile", True, True)),
        "Run the compile phase (maps to src_compile).")

    @observer.decorate_build_method("install")
    def install(self):
        """Run the install phase (maps to src_install)."""
        # TODO: replace print() usage with observer
        print(f">>> Install {self.env['PF']} into {self.ED!r} category {self.env['CATEGORY']}")
        ret = self._generic_phase("install", False, True)
        print(f">>> Completed installing {self.env['PF']} into {self.ED!r}")
        return ret

    @observer.decorate_build_method("test")
    def test(self):
        """Run the test phase (if enabled), maps to src_test."""
        if not self.run_test:
            return True
        return self._generic_phase(
            "test", True, True, failure_allowed=self.allow_failed_test)

    def finalize(self):
        """Finalize the operation.

        This yields a built package, but the packages metadata/contents are
        bound to the workdir. In other words, install the package somewhere
        prior to executing clean if you intend on installing it.

        :return: :obj:`pkgcore.ebuild.ebuild_built.package` instance
        """
        factory = ebuild_built.fake_package_factory(self._built_class)
        return factory.new_package(
            self.pkg, self.env["D"], pjoin(self.env["T"], "environment"))


class binpkg_localize(ebd, setup_mixin, format.build):

    stage_depends = {"finalize": "setup", "setup": "start"}
    setup_is_for_src = False

    def __init__(self, domain, pkg, **kwargs):
        self._built_class = ebuild_built.package
        format.build.__init__(self, domain, pkg, {}, observer=kwargs.get("observer", None))
        ebd.__init__(self, pkg, **kwargs)
        if self.eapi.options.has_merge_type:
            self.env["MERGE_TYPE"] = "binpkg"

    def finalize(self):
        return MutatedPkg(self.pkg, {"environment": self.get_env_source()})


class ebuild_operations:

    _checks = []

    def _register_check(checks):
        """Decorator to register sanity checks that will be run."""
        def _wrap_func(func):
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)
            checks.append(func)
            return wrapped
        return _wrap_func

    def _cmd_implementation_sanity_check(self, domain):
        """Run all defined sanity checks."""
        failures = []
        for check in self._checks:
            if result := check(self, self.pkg, domain=domain):
                failures.append(result)
        return failures

    @_register_check(_checks)
    def _check_required_use(self, pkg, **kwargs):
        """Perform REQUIRED_USE verification against a set of USE flags.

        Note that this assumes the REQUIRED_USE depset has been evaluated
        against a known set of enabled USE flags and is in collapsed form.
        """
        if pkg.eapi.options.has_required_use:
            if failures := tuple(node for node in pkg.required_use if not node.match(pkg.use)):
                return errors.RequiredUseError(pkg, failures)

    @_register_check(_checks)
    def _check_pkg_pretend(self, pkg, *, domain, **kwargs):
        """Run pkg_pretend phase."""
        # pkg_pretend is not defined or required
        if 'pretend' not in pkg.mandatory_phases:
            return

        commands = None
        if not pkg.built:
            commands = {
                'request_inherit': partial(inherit_handler, self._eclass_cache),
                'has_version': ebd_ipc.Has_Version(self),
                'best_version': ebd_ipc.Best_Version(self),
            }

        # Use base build tempdir for $T instead of full pkg specific path to
        # avoid having to create/remove directories -- pkg_pretend isn't
        # allowed to write to the filesystem anyway.
        self.env = expected_ebuild_env(pkg)
        self.env["T"] = domain.pm_tmpdir
        ebd.set_path_vars(self.env, pkg, domain)
        # avoid clipping eend() messages
        self.env["PKGCORE_RC_PREFIX"] = '2'

        with TemporaryFile() as f:
            # suppress bash output by default
            fd_pipes = {1: f.fileno(), 2: f.fileno()}
            try:
                run_generic_phase(
                    pkg, "pretend", self.env, tmpdir=None, fd_pipes=fd_pipes,
                    userpriv=True, sandbox=True, extra_handlers=commands)
            except ProcessorError as e:
                f.seek(0)
                output = f.read().decode().strip('\n')
                return errors.PkgPretendError(pkg, output, e)


class src_operations(ebuild_operations, format.build_operations):

    def __init__(self, domain, pkg, eclass_cache, observer=None):
        format.build_operations.__init__(self, domain, pkg, observer=observer)
        self._eclass_cache = eclass_cache

    def _cmd_implementation_build(self, observer, verified_files,
                                  clean=False, force_test=False):
        return buildable(
            self.domain, self.pkg, verified_files,
            self._eclass_cache, observer=observer,
            clean=clean, force_test=force_test)


class misc_operations(ebd):

    def __init__(self, domain, *args, **kwds):
        self.domain = domain
        super().__init__(*args, **kwds)

    def configure(self, observer=None):
        return self._generic_phase('config', False, True)

    def info(self, observer=None):
        return self._generic_phase('info', True, True)


class built_operations(ebuild_operations, format.operations):

    def __init__(self, domain, pkg, observer=None, initial_env=None):
        format.operations.__init__(self, domain, pkg, observer=observer)
        self._initial_env = initial_env
        self._localized_ebd = None

    def _cmd_implementation_localize(self, observer, force=False):
        if not force and getattr(self.pkg, '_is_from_source', False):
            return self.pkg
        self._localized_ebd = op = binpkg_localize(
            self.domain, self.pkg, clean=False,
            initial_env=self._initial_env, env_data_source=self.pkg.environment,
            observer=observer)
        return op.finalize()

    def _cmd_implementation_cleanup(self, observer, force=False):
        if not self._localized_ebd:
            return True
        return self._localized_ebd.cleanup(force=force)

    def _cmd_check_support_configure(self):
        pkg = self.pkg
        if 'config' not in pkg.mandatory_phases:
            return False
        return True

    def _cmd_implementation_configure(self, observer):
        misc = misc_operations(
            self.domain, self.pkg, env_data_source=self.pkg.environment, clean=True)
        try:
            misc.start()
            misc.configure()
        finally:
            misc.cleanup()
        return True



from collections import OrderedDict
from functools import partial

import pytest
from pkgcore.ebuild.atom import atom
from pkgcore.ebuild.cpv import VersionedCPV
from pkgcore.operations.repo import operations
from pkgcore.package.mutated import MutatedPkg
from pkgcore.repository.util import SimpleTree
from pkgcore.restrictions import boolean, packages, values
from pkgcore.test import malleable_obj
from snakeoil.currying import post_curry


class TestPrototype:

    def setup_method(self):
        # we use an OrderedDict here specifically to trigger any sorter
        # related bugs
        d = {
            "dev-util": {"diffball": ["1.0", "0.7"], "bsdiff": ["0.4.1", "0.4.2"]},
            "dev-lib": {"fake": ["1.0", "1.0-r1"]}}
        self.repo = SimpleTree(
            OrderedDict((k, d[k]) for k in sorted(d, reverse=True)))

    def test_concurrent_access(self):
        iall = iter(self.repo)
        self.repo.match(atom("dev-lib/fake"))
        pkg = next(iall)
        if pkg.category == 'dev-util':
            self.repo.match(atom("dev-lib/fake"))
        else:
            self.repo.match(atom("dev-util/diffball"))
        # should not explode...
        list(iall)

    def test_internal_lookups(self):
        assert sorted(self.repo.categories) == sorted(["dev-lib", "dev-util"])
        assert \
            sorted(map("/".join, self.repo.versions)) == \
            sorted([x for x in ["dev-util/diffball", "dev-util/bsdiff", "dev-lib/fake"]])
        assert \
            sorted(
                f"{cp[0]}/{cp[1]}-{v}"
                for cp, t in self.repo.versions.items() for v in t) == \
            sorted([
                "dev-util/diffball-1.0", "dev-util/diffball-0.7",
                "dev-util/bsdiff-0.4.1", "dev-util/bsdiff-0.4.2",
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"])

    def test_simple_query(self):
        a = atom("=dev-util/diffball-1.0")
        self.repo.match(a)
        assert self.repo.match(a)
        assert not self.repo.match(atom("dev-util/monkeys_rule"))

    def test_identify_candidates(self):
        with pytest.raises(TypeError):
            self.repo.match("asdf")
        rc = packages.PackageRestriction(
            "category", values.StrExactMatch("dev-util"))
        assert \
            sorted(set(x.package for x in self.repo.itermatch(rc))) == \
            sorted(["diffball", "bsdiff"])
        rp = packages.PackageRestriction(
            "package", values.StrExactMatch("diffball"))
        assert list(x.version for x in self.repo.itermatch(rp, sorter=sorted)) == ["0.7", "1.0"]
        assert \
            self.repo.match(packages.OrRestriction(rc, rp), sorter=sorted) == \
            sorted(VersionedCPV(x) for x in (
                "dev-util/diffball-0.7", "dev-util/diffball-1.0",
                "dev-util/bsdiff-0.4.1", "dev-util/bsdiff-0.4.2"))
        assert \
            sorted(self.repo.itermatch(packages.AndRestriction(rc, rp))) == \
            sorted(VersionedCPV(x) for x in (
                "dev-util/diffball-0.7", "dev-util/diffball-1.0"))
        assert sorted(self.repo) == self.repo.match(packages.AlwaysTrue, sorter=sorted)
        # mix/match cat/pkg to check that it handles that corner case
        # properly for sorting.
        assert \
            sorted(self.repo, reverse=True) == \
            self.repo.match(packages.OrRestriction(
                rc, rp, packages.AlwaysTrue),
                sorter=partial(sorted, reverse=True))
        rc2 = packages.PackageRestriction(
            "category", values.StrExactMatch("dev-lib"))
        assert sorted(self.repo.itermatch(packages.AndRestriction(rp, rc2))) == []

        # note this mixes a category level match, and a pkg level
        # match. they *must* be treated as an or.
        assert \
            sorted(self.repo.itermatch(packages.OrRestriction(rp, rc2))) == \
            sorted(VersionedCPV(x) for x in (
                "dev-util/diffball-0.7", "dev-util/diffball-1.0",
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        # this is similar to the test above, but mixes a cat/pkg
        # candidate with a pkg candidate
        rp2 = packages.PackageRestriction(
            "package", values.StrExactMatch("fake"))
        r = packages.OrRestriction(atom("dev-util/diffball"), rp2)
        assert \
            sorted(self.repo.itermatch(r)) == \
            sorted(VersionedCPV(x) for x in (
                "dev-util/diffball-0.7", "dev-util/diffball-1.0",
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        assert \
            sorted(self.repo.itermatch(
                packages.OrRestriction(packages.AlwaysTrue, rp2))) == \
            sorted(VersionedCPV(x) for x in (
                "dev-util/diffball-0.7", "dev-util/diffball-1.0",
                "dev-util/bsdiff-0.4.1", "dev-util/bsdiff-0.4.2",
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        assert \
            sorted(self.repo.itermatch(packages.PackageRestriction(
                'category', values.StrExactMatch('dev-util', negate=True)))) == \
            sorted(VersionedCPV(x) for x in ("dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        obj = malleable_obj(livefs=False)
        pkg_cls = post_curry(MutatedPkg, {'repo': obj})
        assert \
            sorted(self.repo.itermatch(boolean.AndRestriction(boolean.OrRestriction(
                packages.PackageRestriction(
                    "repo.livefs", values.EqualityMatch(False)),
                packages.PackageRestriction(
                    "category", values.StrExactMatch("virtual"))),
                atom("dev-lib/fake")),
                pkg_cls=pkg_cls)) == \
            sorted(VersionedCPV(x) for x in (
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        assert \
            sorted(self.repo.itermatch(packages.PackageRestriction(
                'category', values.StrExactMatch('dev-lib', negate=True),
                negate=True))) == \
            sorted(VersionedCPV(x) for x in (
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

        assert \
            sorted(self.repo.itermatch(packages.PackageRestriction(
                'category', values.StrExactMatch('dev-lib', negate=True), negate=True))) == \
            sorted(VersionedCPV(x) for x in (
                "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))

    def test_iter(self):
        expected = sorted(VersionedCPV(x) for x in (
            "dev-util/diffball-1.0", "dev-util/diffball-0.7",
            "dev-util/bsdiff-0.4.1", "dev-util/bsdiff-0.4.2",
            "dev-lib/fake-1.0", "dev-lib/fake-1.0-r1"))
        assert sorted(self.repo) == expected

    def test_notify_remove(self):
        pkg = VersionedCPV("dev-util/diffball-1.0")
        self.repo.notify_remove_package(pkg)
        assert list(self.repo.versions[(pkg.category, pkg.package)]) == ["0.7"]

        # test version being emptied, and package updated
        pkg = VersionedCPV("dev-util/diffball-0.7")
        self.repo.notify_remove_package(pkg)
        assert (pkg.category, pkg.package) not in self.repo.versions
        assert pkg.package not in self.repo.packages[pkg.category]

        # test no remaining packages, category updated
        pkg = VersionedCPV("dev-util/bsdiff-0.4.1")
        self.repo.notify_remove_package(pkg)

        pkg = VersionedCPV("dev-util/bsdiff-0.4.2")
        self.repo.notify_remove_package(pkg)
        assert (pkg.category, pkg.package) not in self.repo.versions
        assert pkg.category not in self.repo.packages
        assert pkg.category not in self.repo.categories

    def test_notify_add(self):
        pkg = VersionedCPV("dev-util/diffball-1.2")
        self.repo.notify_add_package(pkg)
        assert sorted(self.repo.versions[(pkg.category, pkg.package)]) == \
            sorted(["1.0", "1.2", "0.7"])

        pkg = VersionedCPV("foo/bar-1.0")
        self.repo.notify_add_package(pkg)
        assert pkg.category in self.repo.categories
        assert pkg.category in self.repo.packages
        ver_key = (pkg.category, pkg.package)
        assert ver_key in self.repo.versions
        assert list(self.repo.versions[ver_key]) == ["1.0"]

        pkg = VersionedCPV("foo/cows-1.0")
        self.repo.notify_add_package(pkg)
        assert (pkg.category, pkg.package) in self.repo.versions

    def _simple_redirect_test(self, attr, arg1='=dev-util/diffball-1.0', arg2=None):
        l = []
        uniq_obj = object()

        def f(*a, **kw):
            a = a[1:-1]
            l.extend((a, kw))
            return uniq_obj
        # if replace, override _replace since replace reflects to it

        class my_ops(operations):
            locals()[f'_cmd_implementation_{attr}'] = f
        self.repo.operations_kls = my_ops
        args = [self.repo.match(atom(arg1))]
        if arg2:
            args.append(VersionedCPV(arg2))
        self.repo.frozen = False
        op = getattr(self.repo.operations, attr)

        def simple_check(op, args, **kw):
            l[:] = []
            assert op(*args, **kw) == uniq_obj
            assert len(l) == 2
            assert list(l[0]) == args
            assert l

        assert self.repo.operations.supports(attr)
        simple_check(op, args)
        assert not l[1]
        simple_check(op, args)
        assert 'force' not in l[1]
        self.repo.frozen = True
        assert not self.repo.operations.supports(attr)
        assert not hasattr(self.repo.operations, attr)

    test_replace = post_curry(_simple_redirect_test, 'replace', arg2='dev-util/diffball-1.1')
    test_uninstall = post_curry(_simple_redirect_test, 'uninstall')
    test_install = post_curry(_simple_redirect_test, 'install')



__all__ = ("rsync_syncer", "rsync_timestamp_syncer",)

import os
import socket
import tempfile
import time

from snakeoil.osutils import pjoin

from ..config.hint import ConfigHint
from . import base


class rsync_syncer(base.ExternalSyncer):

    default_excludes = ['/distfiles', '/local', '/packages']
    default_includes = []
    default_conn_timeout = 15
    default_opts = [
        '--recursive',
        '--delete',
        '--delete-delay',
        '--perms',
        '--times',
        '--compress',
        '--force',
        '--links',
        '--safe-links',
        '--stats',
        '--human-readable',
        '--timeout=180',
        '--whole-file', # this one probably shouldn't be a default
    ]

    default_retries = 5
    binary = "rsync"

    @classmethod
    def _parse_uri(cls, raw_uri):
        if not raw_uri.startswith("rsync://") and \
                not raw_uri.startswith("rsync+"):
            raise base.UriError(raw_uri, "doesn't start with rsync:// nor rsync+")

        if raw_uri.startswith("rsync://"):
            return None, raw_uri

        proto = raw_uri.split(":", 1)
        proto[0] = proto[0].split("+", 1)[1]
        cls.require_binary(proto[0])
        return proto[0], f"rsync:{proto[1]}"

    pkgcore_config_type = ConfigHint({
        'basedir': 'str', 'uri': 'str', 'conn_timeout': 'str', 'usersync': 'bool',
        'compress': 'bool', 'excludes': 'list', 'includes': 'list',
        'retries': 'str', 'opts': 'list', 'extra_opts': 'list', 'proxy': 'str'},
        typename='syncer')

    def __init__(self, basedir, uri, conn_timeout=default_conn_timeout,
                 usersync=False, compress=False, excludes=(), includes=(),
                 retries=default_retries, proxy=None,
                 opts=(), extra_opts=()):
        uri = uri.rstrip(os.path.sep) + os.path.sep
        self.rsh, uri = self._parse_uri(uri)
        super().__init__(basedir, uri, default_verbosity=1, usersync=usersync)
        self.hostname = self.parse_hostname(self.uri)
        if self.rsh:
            self.rsh = self.require_binary(self.rsh)
        self.opts = list(opts) if opts else list(self.default_opts)
        self.opts.extend(extra_opts)
        if compress:
            self.opts.append("--compress")
        self.opts.append("--contimeout=%i" % int(conn_timeout))
        self.excludes = list(self.default_excludes) + list(excludes)
        self.includes = list(self.default_includes) + list(includes)
        self.retries = int(retries)
        self.use_proxy = proxy is not None
        if self.use_proxy:
            self.env['RSYNC_PROXY'] = proxy
        self.is_ipv6 = "--ipv6" in self.opts or "-6" in self.opts
        self.is_ipv6 = self.is_ipv6 and socket.has_ipv6

    @staticmethod
    def parse_hostname(uri):
        return uri[len("rsync://"):].split("@", 1)[-1].split("/", 1)[0]

    def _get_ips(self):
        if self.use_proxy:
            # If we're using a proxy, name resolution is best left to the proxy.
            yield self.hostname
            return

        af_fam = socket.AF_INET
        if self.is_ipv6:
            af_fam = socket.AF_INET6
        try:
            for ipaddr in socket.getaddrinfo(
                    self.hostname, None, af_fam, socket.SOCK_STREAM):
                if ipaddr[0] == socket.AF_INET6:
                    yield f"[{ipaddr[4][0]}]"
                else:
                    yield ipaddr[4][0]
        except OSError as e:
            raise base.SyncError(
                f"DNS resolution failed for {self.hostname!r}: {e.strerror}")

    def _sync(self, verbosity, output_fd):
        fd_pipes = {1: output_fd, 2: output_fd}
        opts = list(self.opts)
        if self.rsh:
            opts.append("-e")
            opts.append(self.rsh)
        opts.extend(f"--exclude={x}" for x in self.excludes)
        opts.extend(f"--include={x}" for x in self.includes)
        if verbosity < 0:
            opts.append("--quiet")
        elif verbosity > 0:
            opts.extend('-v' for x in range(verbosity))

        # zip limits to the shortest iterable
        ret = None
        for count, ip in zip(range(self.retries), self._get_ips()):
            cmd = [self.binary_path,
                 self.uri.replace(self.hostname, ip, 1),
                 self.basedir] + opts

            ret = self._spawn(cmd, fd_pipes)
            if ret == 0:
                return True
            elif ret == 1:
                raise base.SyncError("rsync command syntax error: {' '.join(cmd)}")
            elif ret == 11:
                raise base.SyncError("rsync ran out of disk space")
           # need to do something here instead of just restarting...
           # else:
           #     print(ret)
        raise base.SyncError("all attempts failed")


class _RsyncFileSyncer(rsync_syncer):
    """Support syncing a single file over rsync."""

    def __init__(self, path, uri):
        super().__init__(basedir=path, uri=uri)
        # override parent classes that always assume directory syncing
        self.basedir = path
        self.uri = uri


class rsync_timestamp_syncer(rsync_syncer):

    forcable = True
    forward_sync_delay = 25 * 60 # 25 minutes
    negative_sync_delay = 60 * 60 # 60 minutes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_timestamp = self.current_timestamp()

    def current_timestamp(self, path=None):
        """
        :param path: override the default path for the timestamp to read
        :return: string of the timestamp data
        """
        if path is None:
            path = pjoin(self.basedir, "metadata", "timestamp.chk")
        try:
            with open(path) as f:
                date, offset = f.read().strip().rsplit('+', 1)
            date = time.mktime(time.strptime(date, "%a, %d %b %Y %H:%M:%S "))
            # add the hour/minute offset
            date += int(offset[:2] * 60) + int(offset[2:])
            return date
        except (FileNotFoundError, NotADirectoryError):
            return None
        except ValueError:
            # malformed timestamp
            return None

    def _sync(self, verbosity, output_fd, force=False):
        doit = force or self.last_timestamp is None
        ret = None
        try:
            if not doit:
                # try to sync the timestamp file to check the delta
                with tempfile.NamedTemporaryFile() as new_timestamp:
                    timestamp_uri = pjoin(self.uri, "metadata", "timestamp.chk")
                    timestamp_path = new_timestamp.name
                    timestamp_syncer = _RsyncFileSyncer(timestamp_path, timestamp_uri)
                    ret = timestamp_syncer._sync(verbosity, output_fd)
                    if not ret:
                        doit = True
                    else:
                        delta = self.current_timestamp(timestamp_path) - self.last_timestamp
                        if delta >= 0:
                            doit = delta > self.forward_sync_delay
                        else:
                            doit = delta > self.negative_sync_delay
            if not doit:
                return True
            ret = super()._sync(verbosity, output_fd)
            # force a reset of the timestamp
            self.last_timestamp = self.current_timestamp()
        finally:
            if ret:
                return ret
            # ensure the timestamp is back to the old
            try:
                timestamp_path = pjoin(self.basedir, "metadata", "timestamp.chk")
                if self.last_timestamp is None:
                    os.remove(timestamp_path)
                else:
                    with open(timestamp_path, "w") as f:
                        f.write(time.strftime("%a, %d %b %Y %H:%M:%S +0000",
                                time.gmtime(self.last_timestamp)))
            except EnvironmentError:
                # don't care...
                pass
        return ret



from unittest import mock

import pytest
from pkgcore.sync import base, cvs
from snakeoil.process import CommandNotFound


class TestCVSSyncer:

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = tmp_path / 'repo'

    def test_uri_parse(self):
        # external binary doesn't exist
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.side_effect = CommandNotFound('cvs')
            with pytest.raises(base.SyncError):
                cvs.cvs_syncer(
                    str(self.repo_path), "cvs+/bin/sh://foon.com/dar")

        # fake that the external binary exists
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.return_value = 'cvs'

            # nonexistent rsh
            with mock.patch('pkgcore.sync.base.ExternalSyncer.require_binary') as require_binary:
                require_binary.side_effect = base.MissingBinary('', 'rsh')
                with pytest.raises(base.SyncError):
                    cvs.cvs_syncer(str(self.repo_path), "cvs+rsh://foon.com/dar")

            o = cvs.cvs_syncer(str(self.repo_path), "cvs://dar:module")
            assert o.uri == ":anoncvs:dar"
            assert o.module == "module"
            assert o.rsh == None
            assert o.env["CVSROOT"] == ":anoncvs:dar"

            o = cvs.cvs_syncer(str(self.repo_path), "cvs+pserver://dar:module")
            assert o.uri == ":pserver:dar"
            assert o.module == "module"
            assert o.rsh == None
            assert o.env["CVSROOT"] == ":pserver:dar"

            with mock.patch('pkgcore.sync.base.ExternalSyncer.require_binary') as require_binary:
                require_binary.return_value = '/bin/sh'
                o = cvs.cvs_syncer(str(self.repo_path), "cvs+/bin/sh://dar:module")
                assert o.rsh == "/bin/sh"
                assert o.uri == ":ext:dar"
                assert o.env["CVSROOT"] == ":ext:dar"
                assert o.env["CVS_RSH"] == "/bin/sh"



"""low-level ebuild utility"""

import sys

from snakeoil.cli.exceptions import ExitException
from snakeoil.strings import pluralism

from ..operations import OperationError, observer
from ..package.errors import MetadataException
from ..util.commandline import ArgumentParser, StoreTarget

argparser = ArgumentParser(description=__doc__, script=(__file__, __name__))
argparser.add_argument(
    'target', action=StoreTarget,
    allow_ebuild_paths=True, allow_external_repos=True,
    help="atom or ebuild matching a pkg to execute phases from")
argparser.add_argument('phase', nargs='+', help="phases to run")
phase_opts = argparser.add_argument_group("phase options")
phase_opts.add_argument(
    "--no-auto", action='store_true', default=False,
    help="run just the specified phases; "
         "it's up to the invoker to get the order right")


@argparser.bind_final_check
def _validate_args(parser, namespace):
    namespace.repo = namespace.domain.ebuild_repos_unfiltered


@argparser.bind_main_func
def main(options, out, err):
    token, restriction = options.target[0]
    domain = options.domain

    try:
        pkgs = options.repo.match(restriction, pkg_filter=None)
    except MetadataException as e:
        error = e.msg(verbosity=options.verbosity)
        argparser.error(f'{e.pkg.cpvstr}::{e.pkg.repo.repo_id}: {error}')

    if not pkgs:
        argparser.error(f"no matches: {token!r}")

    pkg = max(pkgs)
    if len(pkgs) > 1:
        argparser.err.write(f"got multiple matches for {token!r}:")
        if len(set((p.slot, p.repo) for p in pkgs)) != 1:
            for p in pkgs:
                repo_id = getattr(p.repo, 'repo_id', 'unknown')
                argparser.err.write(f"{p.cpvstr}:{p.slot}::{repo_id}", prefix='  ')
            argparser.err.write()
            argparser.error("please refine your restriction to one match")
        repo_id = getattr(pkg.repo, 'repo_id', 'unknown')
        argparser.err.write(f"choosing {pkg.cpvstr}:{pkg.slot}::{repo_id}", prefix='  ')
        sys.stderr.flush()

    kwds = {}
    phase_obs = observer.phase_observer(observer.formatter_output(out), options.debug)

    phases = [x for x in options.phase if x != 'clean']
    clean = (len(phases) != len(options.phase))

    if options.no_auto:
        kwds["ignore_deps"] = True
        if "setup" in phases:
            phases.insert(0, "fetch")

    # forcibly run test phase if selected
    force_test = 'test' in phases
    if force_test and 'test' in pkg.iuse:
        pkg.use.add('test')

    # by default turn off startup cleans; we clean by ourselves if
    # told to do so via an arg
    build = domain.build_pkg(
        pkg, failed=True, clean=False, observer=phase_obs, force_test=force_test)
    if clean:
        build.cleanup(force=True)
    build._reload_state()

    phase_funcs = [(p, getattr(build, p, None)) for p in phases]
    unknown_phases = [p for p, func in phase_funcs if func is None]
    if unknown_phases:
        argparser.error("unknown phase%s: %s" % (
            pluralism(unknown_phases), ', '.join(map(repr, unknown_phases))))

    try:
        for phase, func in phase_funcs:
            out.write(f'executing phase {phase}')
            func(**kwds)
    except OperationError as e:
        raise ExitException(f"caught exception executing phase {phase}: {e}") from e



"""
functionality to hand off to a callable, enabling collapsing
long chains of restrictions into Nlog N, or generating
restrictions on the fly
"""

__all__ = ("delegate",)

from . import restriction


class delegate(restriction.base):
    """hand off matching to a handed in prototype

    Example usage of this class should be available in
    :obj:`pkgcore.ebuild.domain`.
    """

    __slots__ = ('_transform', 'negate')

    type = restriction.package_type
    inst_caching = False

    def __init__(self, transform_func, negate=False):
        """
        :param transform_func: callable invoked with data, pkg, and mode
            mode may be "match", "force_True", or "force_False"
        """

        if not callable(transform_func):
            raise TypeError(transform_func)

        object.__setattr__(self, "negate", negate)
        object.__setattr__(self, "_transform", transform_func)

    def match(self, pkginst):
        return self._transform(pkginst, "match") != self.negate

    def force_True(self, pkginst):
        if self.negate:
            return self._transform(pkginst, "force_False")
        return self._transform(pkginst, "force_True")

    def force_False(self, pkginst):
        if self.negate:
            return self._transform(pkginst, "force_True")
        return self._transform(pkginst, "force_False")



__all__ = ("darcs_syncer",)

from . import base


class darcs_syncer(base.VcsSyncer):

    binary = "darcs"

    supported_uris = (
        ('darcs+', 5),
    )

    @staticmethod
    def parse_uri(raw_uri):
        if not raw_uri.startswith("darcs+"):
            raise base.UriError(raw_uri, "doesn't start with darcs+")
        return raw_uri[6:]

    def _initial_pull(self):
        return [self.binary_path, "clone", self.uri, self.basedir]

    def _update_existing(self):
        return [self.binary_path, "pull", self.uri]



from functools import partial
from io import BytesIO

from pkgcore.config import basics
from pkgcore.config.hint import ConfigHint, configurable
from pkgcore.ebuild.cpv import CPV
from pkgcore.operations.repo import install, operations, replace, uninstall
from pkgcore.repository import syncable, util
from pkgcore.scripts import pmaint
from pkgcore.sync import base
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.formatters import PlainTextFormatter
from snakeoil.mappings import AttrAccessible
from snakeoil.test import TestCase

Options = AttrAccessible


class fake_operations(operations):

    def _cmd_implementation_install(self, pkg, observer):
        self.repo.installed.append(pkg)
        return derive_op('add_data', install, self.repo, pkg, observer)

    def _cmd_implementation_uninstall(self, pkg, observer):
        self.repo.uninstalled.append(pkg)
        return derive_op('remove_data', uninstall, self.repo, pkg, observer)

    def _cmd_implementation_replace(self, oldpkg, newpkg, observer):
        self.repo.replaced.append((oldpkg, newpkg))
        return derive_op(('add_data', 'remove_data'),
            replace, self.repo, oldpkg, newpkg, observer)


class fake_repo(util.SimpleTree):

    operations_kls = fake_operations

    def __init__(self, data, frozen=False, livefs=False, repo_id=None):
        self.installed = []
        self.replaced = []
        self.uninstalled = []
        util.SimpleTree.__init__(
            self, data, pkg_klass=partial(fake_pkg, self), repo_id=repo_id)
        self.livefs = livefs
        self.frozen = frozen


def make_repo_config(repo_data, livefs=False, frozen=False, repo_id=None):
    def repo():
        return fake_repo(repo_data, livefs=livefs, frozen=frozen, repo_id=repo_id)
    repo.pkgcore_config_type = ConfigHint(typename='repo')
    return basics.HardCodedConfigSection({'class':repo})


class FakeDomain:

    pkgcore_config_type = ConfigHint({'repos': 'refs:repo',
                                      'binpkg': 'refs:repo',
                                      'vdb': 'refs:repo'},
                                     typename='domain')

    def __init__(self, repos, binpkg, vdb):
        object.__init__(self)
        self.repos = repos
        self.source_repos_raw = util.RepositoryGroup(repos)
        self.installed_repos = util.RepositoryGroup(vdb)
        self.binary_repos_raw = util.RepositoryGroup(binpkg)
        self.vdb = vdb


def make_domain(repo=None, binpkg=None, vdb=None):
    if repo is None:
        repo = {}
    if binpkg is None:
        binpkg = {}
    if vdb is None:
        vdb = {}
    repos_config = make_repo_config(repo, repo_id='fake')
    binpkg_config = make_repo_config(binpkg, frozen=False, repo_id='fake_binpkg')
    vdb_config = make_repo_config(vdb, repo_id='fake_vdb')

    return basics.HardCodedConfigSection({
        'class': FakeDomain,
        'repos': [repos_config],
        'binpkg': [binpkg_config],
        'vdb': [vdb_config],
        'default': True,
        })


class FakeSyncer(base.Syncer):

    def __init__(self,  *args, **kwargs):
        self.succeed = kwargs.pop('succeed', True)
        super().__init__(*args, **kwargs)
        self.synced = False

    def _sync(self, verbosity, output_fd, **kwds):
        self.synced = True
        return self.succeed


class SyncableRepo(syncable.tree, util.SimpleTree):

    pkgcore_config_type = ConfigHint(typename='repo_config')

    def __init__(self, succeed=True):
        util.SimpleTree.__init__(self, {})
        syncer = FakeSyncer('/fake', 'fake', succeed=succeed)
        syncable.tree.__init__(self, syncer)


success_section = basics.HardCodedConfigSection({'class': SyncableRepo,
                                                 'succeed': True})
failure_section = basics.HardCodedConfigSection({'class': SyncableRepo,
                                                 'succeed': False})


class TestSync(TestCase, ArgParseMixin):

    _argparser = pmaint.sync

    def test_parser(self):
        values = self.parse(repo=success_section)
        self.assertEqual(['repo'], [x[0] for x in values.repos])
        values = self.parse('repo', repo=success_section)
        self.assertEqual(['repo'], [x[0] for x in values.repos])

    def test_sync(self):
        config = self.assertOut(
            [
                "*** syncing myrepo",
                "*** synced myrepo",
                ],
            myrepo=success_section)
        self.assertTrue(config.repo_config['myrepo']._syncer.synced)
        self.assertOut(
            [
                "*** syncing myrepo",
                "!!! failed syncing myrepo",
                ],
            myrepo=failure_section)
        self.assertOutAndErr(
            [
                "*** syncing goodrepo",
                "*** synced goodrepo",
                "*** syncing badrepo",
                "!!! failed syncing badrepo",
                "",
                "*** sync results:",
                "*** synced: goodrepo",
                "!!! failed: badrepo",
                ], [],
            'goodrepo', 'badrepo',
            goodrepo=success_section, badrepo=failure_section)


class fake_pkg(CPV):

    def __init__(self, repo, *a, **kw):
        CPV.__init__(self, *a, **kw)
        object.__setattr__(self, 'repo', repo)

def derive_op(name, op, *a, **kw):
    if isinstance(name, str):
        name = [name]
    name = ['finalize_data'] + list(name)
    class new_op(op):
        def f(*a, **kw):
            return True
        for x in name:
            locals()[x] = f
        del f, x

    return new_op(*a, **kw)


class TestCopy(TestCase, ArgParseMixin):

    _argparser = pmaint.copy

    def execute_main(self, *a, **kw):
        config = self.parse(*a, **kw)
        out = PlainTextFormatter(BytesIO())
        ret = config.main_func(config, out, out)
        return ret, config, out

    def test_normal_function(self):
        ret, config, out = self.execute_main(
            'fake_binpkg', '--source-repo', 'fake_vdb',
            '*',
            domain=make_domain(vdb={'sys-apps':{'portage':['2.1', '2.3']}}),
            )
        self.assertEqual(ret, 0, "expected non zero exit code")
        self.assertEqual(list(map(str, config.target_repo.installed)),
            ['sys-apps/portage-2.1', 'sys-apps/portage-2.3'])
        self.assertEqual(config.target_repo.uninstalled,
            config.target_repo.replaced,
            msg="uninstalled should be the same as replaced; empty")

        d = {'sys-apps':{'portage':['2.1', '2.2']}}
        ret, config, out = self.execute_main(
            'fake_binpkg', '--source-repo', 'fake_vdb',
            '=sys-apps/portage-2.1',
            domain=make_domain(binpkg=d, vdb=d),
            )
        self.assertEqual(ret, 0, "expected non zero exit code")
        self.assertEqual([list(map(str, x)) for x in config.target_repo.replaced],
            [['sys-apps/portage-2.1', 'sys-apps/portage-2.1']])
        self.assertEqual(config.target_repo.uninstalled,
            config.target_repo.installed,
            msg="installed should be the same as uninstalled; empty")

    def test_ignore_existing(self):
        ret, config, out = self.execute_main(
            'fake_binpkg', '--source-repo', 'fake_vdb',
            '*', '--ignore-existing',
            domain=make_domain(vdb={'sys-apps':{'portage':['2.1', '2.3']}}),
            )
        self.assertEqual(ret, 0, "expected non zero exit code")
        self.assertEqual(list(map(str, config.target_repo.installed)),
            ['sys-apps/portage-2.1', 'sys-apps/portage-2.3'])
        self.assertEqual(config.target_repo.uninstalled,
            config.target_repo.replaced,
            msg="uninstalled should be the same as replaced; empty")

        ret, config, out = self.execute_main(
            'fake_binpkg', '--source-repo', 'fake_vdb',
            '*', '--ignore-existing',
            domain=make_domain(
                binpkg={'sys-apps':{'portage':['2.1']}},
                vdb={'sys-apps':{'portage':['2.1', '2.3']}}),
            )
        self.assertEqual(ret, 0, "expected non zero exit code")
        self.assertEqual(list(map(str, config.target_repo.installed)),
            ['sys-apps/portage-2.3'])
        self.assertEqual(config.target_repo.uninstalled,
            config.target_repo.replaced,
            msg="uninstalled should be the same as replaced; empty")


class TestRegen(TestCase, ArgParseMixin):

    _argparser = pmaint.regen

    def test_parser(self):

        options = self.parse(
            'fake', '--threads', '2', domain=make_domain())
        self.assertTrue(isinstance(options.repos[0], util.SimpleTree))
        self.assertEqual(options.threads, 2)



"""
conditional attributes on a package.

Changing them triggering regen of other attributes on the package instance.
"""

__all__ = ("make_wrapper",)

from copy import copy
from functools import partial
from operator import attrgetter

from snakeoil.containers import LimitedChangeSet, Unchangable

from .base import wrapper


def _getattr_wrapped(attr, self):
    o = self._cached_wrapped.get(attr)
    if o is None or o[0] != self._reuse_pt:
        o = self._wrapped_attr[attr](
            getattr(self._raw_pkg, attr),
            self._configurable,
            pkg=self)
        o = self._cached_wrapped[attr] = (self._reuse_pt, o)
    return o[1]


def make_wrapper(wrapped_repo, configurable_attribute_name, attributes_to_wrap=(),
                 kls_injections={}):
    """
    :param configurable_attribute_name: attribute name to add,
        and that is used for evaluating attributes_to_wrap
    :param attributes_to_wrap: mapping of attr_name:callable
        for revaluating the pkg_instance, using the result
        instead of the wrapped pkgs attr.
    """

    if configurable_attribute_name.find(".") != -1:
        raise ValueError("can only wrap first level attributes, "
                         "'obj.dar' fex, not '%s'" %
                         (configurable_attribute_name))

    class PackageWrapper(wrapper):
        """Add a new attribute, and evaluate attributes of a wrapped pkg."""

        __slots__ = (
            "_unchangable", "_configurable", "_reuse_pt",
            "_cached_wrapped", "_disabled", "_domain", "repo",
        )

        _wrapped_attr = attributes_to_wrap
        _configurable_name = configurable_attribute_name
        configurable = True

        def operations(self, domain, **kwds):
            return self._operations(domain, self, **kwds)

        locals()[configurable_attribute_name] = property(attrgetter("_configurable"))

        locals().update(
            (x, property(partial(_getattr_wrapped, x)))
            for x in attributes_to_wrap)

        def __init__(self, pkg_instance, initial_settings=None,
                     disabled_settings=None, unchangable_settings=None):
            """
            :type pkg_instance: :obj:`pkgcore.package.metadata.package`
            :param pkg_instance: instance to wrap.
            :type initial_settings: sequence
            :param initial_settings: initial configuration of the
                configurable_attribute
            :type unchangable_settings: sequence
            :param unchangable_settings: settings that configurable_attribute
                cannot be set to
            """

            if initial_settings is None:
                initial_settings = []
            if disabled_settings is None:
                disabled_settings = []
            if unchangable_settings is None:
                unchangable_settings = []

            sf = object.__setattr__
            sf(self, '_unchangable', unchangable_settings)
            sf(self, '_configurable', LimitedChangeSet(
                initial_settings, unchangable_settings))
            sf(self, '_disabled', disabled_settings)
            sf(self, '_reuse_pt', 0)
            sf(self, 'repo', wrapped_repo)
            sf(self, '_cached_wrapped', {})
            sf(self, '_domain', None)
            super().__init__(pkg_instance)

        def __copy__(self):
            return self.__class__(
                self._raw_pkg, self._configurable_name,
                initial_settings=set(self._configurable),
                disabled_settings=self._disabled,
                unchangable_settings=self._unchangable,
                attributes_to_wrap=self._wrapped_attr)

        def rollback(self, point=0):
            """rollback changes to the configurable attribute to an earlier point

            :param point: must be an int
            """
            self._configurable.rollback(point)
            # yes, nuking objs isn't necessarily required.  easier this way though.
            # XXX: optimization point
            object.__setattr__(self, '_reuse_pt', self._reuse_pt + 1)

        def commit(self):
            """Commit current changes.

            This means that those changes can be reverted from this point out.
            """
            self._configurable.commit()
            object.__setattr__(self, '_reuse_pt',  0)

        def changes_count(self):
            """current commit point for the configurable"""
            return self._configurable.changes_count()

        def request_enable(self, attr, *vals):
            """internal function

            since configurable somewhat steps outside of normal
            restriction protocols, request_enable requests that this
            package instance change its configuration to make the
            restriction return True; if not possible, reverts any changes
            it attempted

            :param attr: attr to try and change
            :param vals: :obj:`pkgcore.restrictions.values.base` instances that
                we're attempting to make match True
            """
            if attr not in self._wrapped_attr:
                if attr == self._configurable_name:
                    entry_point = self.changes_count()
                    try:
                        list(map(self._configurable.add, vals))
                        object.__setattr__(self, '_reuse_pt', self._reuse_pt + 1)
                        return True
                    except Unchangable:
                        self.rollback(entry_point)
                else:
                    a = getattr(self._raw_pkg, attr)
                    for x in vals:
                        if x not in a:
                            break
                    else:
                        return True
                return False
            entry_point = self.changes_count()
            a = getattr(self._raw_pkg, attr)
            try:
                for x in vals:
                    for reqs in a.node_conds.get(x, ()):
                        if reqs.force_True(self, attr, vals):
                            break
                    else:
                        self.rollback(entry_point)
                        return False
            except Unchangable:
                self.rollback(entry_point)
                return False
            object.__setattr__(self, '_reuse_pt', self._reuse_pt + 1)
            return True

        def request_disable(self, attr, *vals):
            """internal function

            since configurable somewhat steps outside of normal
            restriction protocols, request_disable requests that this
            package instance change its configuration to make the
            restriction return False; if not possible, reverts any changes
            it attempted

            :param attr: attr to try and change
            :param vals: :obj:`pkgcore.restrictions.values.base` instances that
                we're attempting to make match False
            """
            if attr not in self._wrapped_attr:
                if attr == self._configurable_name:
                    entry_point = self.changes_count()
                    try:
                        list(map(self._configurable.remove, vals))
                        return True
                    except Unchangable:
                        self.rollback(entry_point)
                else:
                    a = getattr(self._raw_pkg, attr)
                    for x in vals:
                        if x in a:
                            break
                    else:
                        return True
                return False
            entry_point = self.changes_count()
            a = getattr(self._raw_pkg, attr)
            try:
                for x in vals:
                    for reqs in a.node_conds.get(x, ()):
                        if reqs.force_False(self, attr, vals):
                            break
                    else:
                        self.rollback(entry_point)
                        return False
            except Unchangable:
                self.rollback(entry_point)
                return False
            object.__setattr__(self, '_reuse_pt', self._reuse_pt + 1)
            return True

        def __str__(self):
            return "config wrapped(%s): %s" % (self._configurable_name,
                                               self._raw_pkg)

        def __repr__(self):
            return "<%s pkg=%r wrapped=%r @%#8x>" % (
                self.__class__.__name__, self._raw_pkg, self._configurable_name,
                id(self))

        def freeze(self):
            o = copy(self)
            o.lock()
            return o

        def lock(self):
            """
            commit any outstanding changes and lock the configuration.
            """
            self.commit()
            object.__setattr__(self, '_configurable', list(self._configurable))

        if 'operations_callback' in kls_injections:
            _operations = kls_injections.pop("operations_callback")
        locals().update(kls_injections)

    return PackageWrapper



import os

from pkgcore.ebuild import eclass_cache
from snakeoil import data_source
from snakeoil.chksum import LazilyHashedPath
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import TempDirMixin


class FakeEclassCache(eclass_cache.base):

    def __init__(self, path):
        eclass_cache.base.__init__(self, location=path, eclassdir=path)
        self.eclasses = {
            "eclass1":LazilyHashedPath(path, mtime=100),
            "eclass2":LazilyHashedPath(path, mtime=200)}


class TestBase(TestCase):

    def setUp(self):
        self.dir = '/nonexistent/path/'
        self.ec = FakeEclassCache(self.dir)
        self.ec_locs = {x: self.dir for x in ("eclass1", "eclass2")}

    def test_rebuild_eclass_entry(self):
        def assertRebuildResults(result, *raw_data):
            i = iter(raw_data)
            data = [(ec, [('mtime', mtime)]) for ec, mtime in zip(i, i)]
            got = self.ec.rebuild_cache_entry(data)
            self.assertTrue(bool(got) == bool(result),
                msg=f"Expected {result!r} from {raw_data!r}, got {got!r}")

        assertRebuildResults(False, 'eclass3', 100)
        assertRebuildResults(True, 'eclass1', 100)
        assertRebuildResults(False, 'eclass1', 200)

    def test_get_eclass_data(self):
        keys = list(self.ec.eclasses.keys())
        data = self.ec.get_eclass_data([])
        self.assertIdentical(data, self.ec.get_eclass_data([]))
        data = self.ec.get_eclass_data(keys)
        self.assertIdentical(data, self.ec.get_eclass_data(keys))
        self.assertEqual(sorted(keys), sorted(data))
        data = self.ec.get_eclass_data(["eclass1"])
        self.assertEqual(data, {'eclass1':self.ec.eclasses['eclass1']})


class TestEclassCache(TempDirMixin, TestBase):

    def setUp(self):
        TempDirMixin.setUp(self)
        for x, mtime in (("eclass1", 100), ("eclass2", 200)):
            open(pjoin(self.dir, f"{x}.eclass"), "w").close()
            os.utime(pjoin(self.dir, f"{x}.eclass"), (mtime, mtime))
        # insert a crap file to ensure it doesn't grab it.
        open(pjoin(self.dir, 'foon-eclass'), 'w').close()
        self.ec = eclass_cache.cache(self.dir)
        self.ec_locs = {x: self.dir for x in ("eclass1", "eclass2")}

    def test_get_eclass(self):
        for x in ("eclass1", "eclass2"):
            handle = self.ec.get_eclass(x)
            self.assertInstance(handle, data_source.local_source)
            self.assertEqual(pjoin(self.ec_locs[x], f"{x}.eclass"),
                handle.path)

        # note an eclass, thus shouldn't grab it.
        self.assertEqual(None, self.ec.get_eclass("foon"))
        self.assertEqual(None, self.ec.get_eclass("foon-eclass"))


class TestStackedCaches(TestEclassCache):

    def setUp(self):
        TempDirMixin.setUp(self)
        self.loc1 = pjoin(self.dir, "stack1")
        self.loc2 = pjoin(self.dir, "stack2")

        os.mkdir(self.loc1)
        open(pjoin(self.loc1, 'eclass1.eclass'), 'w').close()
        os.utime(pjoin(self.loc1, 'eclass1.eclass'), (100, 100))
        self.ec1 = eclass_cache.cache(self.loc1)

        os.mkdir(self.loc2)
        open(pjoin(self.loc2, 'eclass2.eclass'), 'w').close()
        os.utime(pjoin(self.loc2, 'eclass2.eclass'), (100, 100))
        self.ec2 = eclass_cache.cache(self.loc2)
        self.ec = eclass_cache.StackedCaches([self.ec1, self.ec2])
        self.ec_locs = {"eclass1":self.loc1, "eclass2":self.loc2}
        # make a shadowed file to verify it's not seen
        open(pjoin(self.loc2, 'eclass1.eclass'), 'w').close()
